from __future__ import annotations

import ast
import builtins
import inspect
import io
import json
import os
import re
import sys
import textwrap
import types as py_types
from _ast import AsyncFunctionDef, ExceptHandler
from typing import Any, Callable, TypedDict, get_args, get_origin

from weave.trace import settings
from weave.trace.context.weave_client_context import get_weave_client
from weave.trace.ipython import (
    ClassNotFoundError,
    get_class_source,
    is_running_interactively,
)
from weave.trace.op import Op, as_op, is_op
from weave.trace.refs import ObjectRef
from weave.trace.sanitize import REDACTED_VALUE, should_redact
from weave.trace.serialization import serializer
from weave.trace.serialization.mem_artifact import MemTraceFilesArtifact
from weave.trace_server.trace_server_interface_util import str_digest

WEAVE_OP_PATTERN = re.compile(r"@weave\.op(\(\))?")
WEAVE_OP_NO_PAREN_PATTERN = re.compile(r"@weave\.op(?!\()")

CODE_DEP_ERROR_SENTINEL = "<error>"


def arg_names(args: ast.arguments) -> set[str]:
    arg_names = set()
    for arg in args.args:
        arg_names.add(arg.arg)
    for arg in args.kwonlyargs:
        arg_names.add(arg.arg)
    for arg in args.posonlyargs:
        arg_names.add(arg.arg)
    if args.vararg:
        arg_names.add(args.vararg.arg)
    if args.kwarg:
        arg_names.add(args.kwarg.arg)
    return arg_names


class ExternalVariableFinder(ast.NodeVisitor):
    """Given a an AST of a python function, find all external variable references.

    External variable references are variables that are used in the function but
    not created in the function.

    The general strategy is to walk the AST keeping track of what variables are in
    scope. Any variables loads we encounter that are not in the current scope are
    external.
    """

    def __init__(self) -> None:
        self.external_vars: dict[str, bool] = {}
        self.scope_stack: list[set[str]] = [set()]  # Start with a global scope

    def visit_Import(self, node: ast.Import) -> None:
        for alias in node.names:
            self.scope_stack[-1].add(alias.asname or alias.name)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        for alias in node.names:
            self.scope_stack[-1].add(alias.asname or alias.name)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        # Add function name to the current scope
        self.scope_stack[-1].add(node.name)
        # Add function arguments to new scope
        self.scope_stack.append(arg_names(node.args))
        self.generic_visit(node)
        self.scope_stack.pop()  # Pop function scope when we exit the function

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.scope_stack[-1].add(node.name)
        self.scope_stack.append(arg_names(node.args))
        self.generic_visit(node)
        self.scope_stack.pop()  # Pop function scope when we exit the function

    def visit_Lambda(self, node: ast.Lambda) -> None:
        # Add function arguments to the current scope
        self.scope_stack.append(arg_names(node.args))
        self.generic_visit(node)
        self.scope_stack.pop()  # Pop function scope when we exit the function

    def visit_ListComp(self, node: ast.ListComp) -> None:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.elt)

    def visit_SetComp(self, node: ast.SetComp) -> Any:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.elt)

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> None:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.elt)

    def visit_DictComp(self, node: ast.DictComp) -> None:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.key)
        self.visit(node.value)

    def visit_ExceptHandler(self, node: ExceptHandler) -> None:
        if node.name is None:
            return
        self.scope_stack[-1].add(node.name)
        self.generic_visit(node)

    def visit_Name(self, node: ast.Name) -> None:
        # print("  VISIT NAME", node.id, node.ctx)
        # If a variable is used (loaded) but not defined in any scope in the stack, and not builtin it's external
        # TODO: we don't capture python version, but builtins can change from version to version!
        if isinstance(node.ctx, ast.Store):
            self.scope_stack[-1].add(node.id)
        elif isinstance(node.ctx, ast.Load) and not any(
            node.id in scope for scope in self.scope_stack
        ):
            self.external_vars[node.id] = True


def resolve_var(fn: Callable, var_name: str) -> Any:
    """Given a python function, resolve a non-local variable name."""
    # First to see if the variable is in the closure
    if fn.__closure__:
        closure_vars = {}
        # __code__.co_freevars is the closure variable names in order
        for vn, closure_cell in zip(fn.__code__.co_freevars, fn.__closure__):
            closure_vars[vn] = closure_cell.cell_contents
        if var_name in closure_vars:
            return closure_vars[var_name]
    if var_name in fn.__globals__:
        return fn.__globals__[var_name]
    return None


class RefJSONEncoder(json.JSONEncoder):
    """Json encoder used for convert storage.to_json_with_refs result to python code"""

    SPECIAL_REF_TOKEN = "__WEAVE_REF__"

    def default(self, o: Any) -> Any:
        if isinstance(o, (ObjectRef)):
            ref_code = f"weave.ref('{str(o)}')"

        if ref_code is not None:
            # This will be a quoted json string in the json.dumps result. We put special
            # tokens in so we can remove the quotes in the final result
            return f"{self.SPECIAL_REF_TOKEN}{ref_code}.get(){self.SPECIAL_REF_TOKEN}"
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)


class GetCodeDepsResult(TypedDict):
    import_code: list[str]
    code: list[str]
    warnings: list[str]


def get_source_notebook_safe(fn: Callable) -> str:
    # In ipython, we can't use inspect.getsource on classes defined in the notebook
    if is_running_interactively() and inspect.isclass(fn):
        try:
            src = get_class_source(fn)
        except ClassNotFoundError:
            # Not all times are we using notebook code - for example if a class
            # is defined int he Weave package itself.
            src = inspect.getsource(fn)
    else:
        src = inspect.getsource(fn)
    return textwrap.dedent(src)


def reconstruct_signature(fn: Callable) -> str:
    sig = inspect.signature(fn)
    module = sys.modules[fn.__module__]

    def make_annotation_str(annotation: Any) -> str:
        if isinstance(annotation, str):
            return annotation
        if annotation is inspect.Parameter.empty:
            return ""

        if (origin := get_origin(annotation)) is not None:
            args = get_args(annotation)
            replaced_args = [make_annotation_str(arg) for arg in args]
            replaced_args_str = ", ".join(replaced_args)

            if origin_name := getattr(origin, "__name__", None):
                return f"{origin_name}[{replaced_args_str}]"
            return f"{origin}[{replaced_args_str}]"

        if isinstance(annotation, type):
            # For builtins, just use the name
            if annotation.__module__ == "builtins":
                return annotation.__name__
            # Otherwise, check if the type is imported and use the alias if given
            for name, obj in module.__dict__.items():
                if isinstance(obj, py_types.ModuleType):
                    if annotation.__module__ == obj.__name__:
                        return f"{name}.{annotation.__name__}"
        return str(annotation)

    def quote_default_str(default: Any) -> Any:
        if isinstance(default, str):
            return f'"{default}"'
        return default

    params = []
    for name, param in sig.parameters.items():
        annotation_str = make_annotation_str(param.annotation)
        if param.default is param.empty:
            default = ""
        else:
            default = f" = {quote_default_str(param.default)}"
        params.append(f"{name}: {annotation_str}{default}")

    return_annotation = make_annotation_str(sig.return_annotation)

    sig_str = f"({', '.join(params)})"
    if return_annotation:
        sig_str += f" -> {return_annotation}"

    return sig_str


def get_source_or_fallback(fn: Callable, *, warnings: list[str]) -> str:
    if fn_is_op := is_op(fn):
        op = as_op(fn)
        fn = op.resolve_fn

    if not settings.should_capture_code() or (
        fn_is_op and not op._code_capture_enabled
    ):
        # This digest is kept for op versioning purposes
        digest = str_digest(inspect.getsource(fn))
        return textwrap.dedent(
            f"""
            def func(*args, **kwargs):
                ...  # Code-capture was disabled while saving this op (digest: {digest})
            """
        )

    try:
        return get_source_notebook_safe(fn)
    except OSError:
        pass

    try:
        sig_str = reconstruct_signature(fn)
    except Exception as e:
        warnings.append(f"Failed to reconstruct signature: {e}")
        sig_str = "(*args, **kwargs)"

    func_name = fn.__name__
    missing_code_template = textwrap.dedent(
        f"""
        def {func_name}{sig_str}:
            ... # Code-capture unavailable for this op
        """
    )[1:]  # skip first newline char

    return missing_code_template


def get_code_deps_safe(
    fn: Callable | type,  # A function or a class
    artifact: MemTraceFilesArtifact,
    depth: int = 0,
) -> GetCodeDepsResult:
    """Given a python function, return source code that contains the dependencies of that function.

    This will:
    - include any functions within the same module that are referenced within the function body
    - import any modules that are referenced within the function body or any referenced function bodies

    Current issues (to fix):
    - imported modules may be within the function's package, which doesn't work (any imported modules need
      to be present in the loading code)
    - doesn't serialize python requirements and other necessary system information
    - type annotations are not well handled, and may cause errors
    - refering to @weave.type() objects will cause errors

    Args:
        fn: The Python function to analyze.

    Returns:
        tupe: A tuple containing
            import_code: str, the code that should be included in the generated code to ensure all
                dependencies are available for the function body.
            code: str, the function body code.
            warnings: list[str], any warnings that occurred during the process.
    """
    try:
        return _get_code_deps(fn, artifact, {}, depth)
    except Exception as e:
        print(f"Error getting code deps for {fn}: {e}")
        return {
            "import_code": [],
            "code": [CODE_DEP_ERROR_SENTINEL],
            "warnings": [f"Error getting code dependencies for function {fn}: {e}"],
        }


def _get_code_deps(
    fn: Callable | type,  # A function or a class
    artifact: MemTraceFilesArtifact,
    seen: dict[Callable | type, bool],
    depth: int = 0,
) -> GetCodeDepsResult:
    warnings: list[str] = []
    if depth > 20:
        warnings = [
            "Recursion depth exceeded in get_code_deps, this may indicate circular dependencies, which are not yet handled."
        ]
        return {"import_code": [], "code": [], "warnings": warnings}

    source = get_source_or_fallback(fn, warnings=warnings)
    try:
        parsed = ast.parse(source)
    except SyntaxError:
        warnings.append(f"Could not parse source of function {fn}.")
        return {"import_code": [], "code": [], "warnings": warnings}

    visitor = ExternalVariableFinder()
    visitor.visit(parsed)
    external_vars = list(visitor.external_vars)

    import_code = []
    code = []
    for var_name in external_vars:
        var_value = None
        if isinstance(fn, py_types.FunctionType):
            var_value = resolve_var(fn, var_name)

        try:
            # Some objects throw on equality comparison (like dataframes)
            equivalent = var_value == fn
        except ValueError:
            equivalent = False
        if equivalent:
            # If the variable is the function itself (recursion), we don't need to include it
            continue
        if var_value is None:
            # Try to resolve the variable from the module that the
            # item appears within.
            module = inspect.getmodule(fn)
            var_value = module.__dict__.get(var_name)
        if var_value is None:
            if getattr(builtins, var_name, None):
                # Its a builtin, carry on
                continue
            warnings.append(
                f'Could not resolve var "{var_name}" declared in body of fn {fn}. This op will not be reloadable, but calls to it will be tracked'
            )
        elif isinstance(var_value, py_types.ModuleType):
            import_line = f"import {var_value.__name__}"
            if var_value.__name__ != var_name:
                import_line += f" as {var_name}"
            import_code.append(import_line)
        elif isinstance(var_value, (py_types.FunctionType, type)) or is_op(var_value):
            if var_value.__module__ == fn.__module__:
                if not var_value in seen:
                    seen[var_value] = True
                    result = _get_code_deps(var_value, artifact, seen, depth + 1)
                    fn_warnings = result["warnings"]
                    fn_import_code = result["import_code"]
                    fn_code = result["code"]

                    import_code += fn_import_code

                    code += fn_code
                    code.append(get_source_notebook_safe(var_value))

                    warnings += fn_warnings
            else:
                # For now, if the function is in another module.
                # we just import it. This is ok for libraries, but not
                # if the user has functions declared within their
                # package but outside of the op module.
                if var_value.__module__.split(".")[0] == fn.__module__.split(".")[0]:
                    pass

                if is_op(var_value):
                    warnings.append(
                        f"Cross-module op dependencies are not yet serializable {var_value}"
                    )
                else:
                    import_line = (
                        f"from {var_value.__module__} import {var_value.__name__}"
                    )
                    if var_value.__name__ != var_name:
                        import_line += f"as {var_name}"

                    import_code.append(import_line)

        else:
            # Code saving for Ops using TypedDict and NotRequired was failing on Python 3.9
            # because NotRequired on 3.9 requires using typing_extensions and the implementation
            # doesn't have a __name__ attribute. We now check for _name as a fallback.
            # This came up in the context of persisting an inline custom object deserializer.
            var_value_name = getattr(
                var_value, "__name__", getattr(var_value, "_name", None)
            )
            if (
                var_value_name
                and hasattr(var_value, "__module__")
                and var_value.__module__ != fn.__module__
            ):
                import_line = f"from {var_value.__module__} import {var_value_name}"
                if var_value_name != var_name:
                    import_line += f" as {var_name}"
                import_code.append(import_line)
            else:
                try:
                    if (client := get_weave_client()) is None:
                        raise ValueError("Weave client not found")

                    from weave.trace.serialization.serialize import to_json

                    # Redact sensitive values
                    if should_redact(var_name):
                        json_val = REDACTED_VALUE
                    else:
                        json_val = to_json(var_value, client._project_id(), client)
                except Exception as e:
                    warnings.append(
                        f"Serialization error for value of {var_name} needed by {fn}. Encountered:\n    {e}"
                    )
                else:
                    if should_redact(var_name):
                        json_str = f'"{REDACTED_VALUE}"'
                    else:
                        json_str = json.dumps(json_val, cls=RefJSONEncoder, indent=4)
                    code_paragraph = f"{var_name} = " + json_str + "\n"
                    code_paragraph = code_paragraph.replace(
                        f'"{RefJSONEncoder.SPECIAL_REF_TOKEN}', ""
                    )
                    code_paragraph = code_paragraph.replace(
                        f'{RefJSONEncoder.SPECIAL_REF_TOKEN}"', ""
                    )
                    code.append(code_paragraph)
    return {"import_code": import_code, "code": code, "warnings": warnings}


def find_last_weave_op_function(
    source_code: str,
) -> ast.FunctionDef | ast.AsyncFunctionDef | None:
    """Given a string of python source code, find the last function that is decorated with 'weave.op'."""
    tree = ast.parse(source_code)

    last_function = None

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) or isinstance(node, AsyncFunctionDef):
            for decorator in node.decorator_list:
                # Check if the decorator is 'weave.op'
                if isinstance(decorator, ast.Name) and decorator.id == "weave.op":
                    last_function = node
                    break  # Break the inner loop, continue with the next function
                # Check if the decorator is a call to 'weave.op()'
                elif (
                    isinstance(decorator, ast.Call)
                    and isinstance(decorator.func, ast.Attribute)
                    and decorator.func.attr == "op"
                    and isinstance(decorator.func.value, ast.Name)
                    and decorator.func.value.id == "weave"
                ):
                    last_function = node
                    break  # Break the inner loop, continue with the next function

    return last_function


def dedupe_list(original_list: list[str]) -> list[str]:
    seen = set()
    deduped = []
    for x in original_list:
        if x not in seen:
            deduped.append(x)
            seen.add(x)
    return deduped


def save_instance(obj: Op, artifact: MemTraceFilesArtifact, name: str) -> None:
    result = get_code_deps_safe(obj.resolve_fn, artifact)
    import_code = result["import_code"]
    code = result["code"]
    warnings = result["warnings"]
    if warnings:
        message = f"Warning: Incomplete serialization for op {obj}. This op may not be reloadable"
        for warning in warnings:
            message += "\n  " + warning

    op_function_code = get_source_or_fallback(obj, warnings=warnings)
    if not obj._code_capture_enabled:
        import_code = []
        code = []

    if settings.should_redact_pii():
        from weave.trace.pii_redaction import redact_pii_string

        op_function_code = redact_pii_string(op_function_code)

    if not WEAVE_OP_PATTERN.search(op_function_code):
        op_function_code = "@weave.op()\n" + op_function_code
    else:
        op_function_code = WEAVE_OP_NO_PAREN_PATTERN.sub(
            "@weave.op()", op_function_code
        )
    code.append(op_function_code)

    with artifact.new_file(f"{name}.py") as f:
        assert isinstance(f, io.StringIO)
        import_block = "\n".join(import_code)
        import_lines = ["import weave"] + import_block.split("\n")
        import_lines = dedupe_list(import_lines)
        import_lines = [l for l in import_lines if "weave.api" not in l]
        import_block = "\n".join(import_lines)
        code_block = "\n".join(code)
        f.write(f"{import_block}\n\n{code_block}")


def load_instance(
    artifact: MemTraceFilesArtifact,
    name: str,
) -> Op | None:
    file_name = f"{name}.py"
    module_path = artifact.path(file_name)

    # Python import caching means we can't just import "obj.py" here
    # because serialized ops would be cached at the "obj" key. We include
    # the version in the module name to avoid this. Since version names
    # are content hashes, this is correct.
    #
    art_and_version_dir = module_path[: -(1 + len(file_name))]
    art_dir, version_subdir = art_and_version_dir.rsplit("/", 1)
    module_dir = art_dir
    import_name = (
        version_subdir + "." + ".".join(os.path.splitext(file_name)[0].split("/"))
    )

    sys.path.insert(0, os.path.abspath(module_dir))
    try:
        mod = __import__(import_name, fromlist=[module_dir])
    except Exception as e:
        print("Op loading exception. This might be fine!", e)
        import traceback

        traceback.print_exc()
        return None
    sys.path.pop(0)

    # In the case where the saved op calls another op, we will have multiple
    # ops in the file. The file will look like
    # op1 = weave.ref('...').get()
    # @weave.op()
    # def op2():
    #     op1()
    #
    # We need to get the last declared op in the module. We can't do this
    # simply by iterating module attributes, because order is not guaranteed,
    # so we resort to looking at the source ast.
    last_op_function = find_last_weave_op_function(inspect.getsource(mod))
    if last_op_function is None:
        print(
            f"Unexpected Weave module saved in: {module_path}. No op defs found. All members: {dir(mod)}. {module_dir=} {import_name=}"
        )
        return None

    od: Op = getattr(mod, last_op_function.name)
    return od


def fully_qualified_opname(wrap_fn: Callable) -> str:
    op_module_file = os.path.abspath(inspect.getfile(wrap_fn))
    if op_module_file.endswith(".py"):
        op_module_file = op_module_file[:-3]
    elif op_module_file.endswith(".pyc"):
        op_module_file = op_module_file[:-4]
    return "file://" + op_module_file + "." + wrap_fn.__name__


serializer.register_serializer(Op, save_instance, load_instance, is_op)



import logging
from typing import TYPE_CHECKING, Any, Optional

from pydantic import Field, PrivateAttr

import weave
from weave.flow.scorer import WeaveScorerResult

if TYPE_CHECKING:
    from presidio_analyzer import (
        AnalyzerEngine,
        RecognizerResult,
    )
    from presidio_anonymizer import AnonymizerEngine


logger = logging.getLogger(__name__)


class PresidioScorer(weave.Scorer):
    """
    The `PresidioScorer` class is a guardrail for entity recognition and anonymization
    by leveraging Presidio's AnalyzerEngine and AnonymizerEngine to perform these tasks.

    Attributes:
        selected_entities (list[str]): A list of entity types to detect in the text.
        language (str): The language of the text to be analyzed.
        custom_recognizers (list[EntityRecognizer]): A list of custom recognizers to add to the
            analyzer that are of type `presidio.EntityRecognizer`.

    Offline mode for presidio: https://github.com/microsoft/presidio/discussions/1435
    """

    language: str = Field(
        default="en", description="The language of the text to be analyzed."
    )
    custom_recognizers: list[Any] = Field(
        default_factory=list,
        description="A list of custom recognizers to add to the analyzer. Check Presidio's documentation for more information; https://microsoft.github.io/presidio/samples/python/customizing_presidio_analyzer/",
    )

    selected_entities: Optional[list[str]] = Field(
        default=None,
        description="A list of entity types to detect in the text.",
        examples=[["EMAIL_ADDRESS"]],
    )

    # Private attributes
    _analyzer: Optional["AnalyzerEngine"] = PrivateAttr(default=None)
    _anonymizer: Optional["AnonymizerEngine"] = PrivateAttr(default=None)

    def model_post_init(self, __context: Any) -> None:
        from presidio_analyzer import AnalyzerEngine, RecognizerRegistry
        from presidio_anonymizer import AnonymizerEngine

        registry = RecognizerRegistry()
        self._analyzer = AnalyzerEngine(registry=registry)
        self._anonymizer = AnonymizerEngine()

        # Add custom recognizers if provided
        if self.custom_recognizers:
            for recognizer in self.custom_recognizers:
                self._analyzer.registry.add_recognizer(recognizer)

        # Get available entities dynamically
        available_entities = [
            recognizer.supported_entities[0]
            for recognizer in self._analyzer.registry.recognizers
        ]

        if self.selected_entities is not None:
            # Filter out invalid entities and warn user
            invalid_entities = list(
                set(self.selected_entities) - set(available_entities)
            )
            valid_entities = list(
                set(self.selected_entities).intersection(available_entities)
            )

            if invalid_entities:
                logger.warning(
                    f"\nThe following entities are not available and will be ignored: {invalid_entities}\nContinuing with valid entities: {valid_entities}"
                )
                self.selected_entities = valid_entities
        else:
            self.selected_entities = available_entities

    @weave.op
    def group_analyzer_results_by_entity_type(
        self, output: str, analyzer_results: list["RecognizerResult"]
    ) -> dict[str, list[str]]:
        """Group results by entity type"""
        detected_entities: dict[str, list[str]] = {}
        for result in analyzer_results:
            entity_type = result.entity_type
            text_chunk = output[result.start : result.end]
            if entity_type not in detected_entities:
                detected_entities[entity_type] = []
            detected_entities[entity_type].append(text_chunk)
        return detected_entities

    @weave.op
    def create_reason(self, detected_entities: dict[str, list[str]]) -> str:
        """Create explanation for why the text was flagged"""
        explanation_parts = []
        if detected_entities:
            explanation_parts.append("Found the following entities in the text:")
            for entity_type, instances in detected_entities.items():
                explanation_parts.append(
                    f"- {entity_type}: {len(instances)} instance(s)"
                )
        else:
            explanation_parts.append("No entities detected in the text.")

        # Add information about what was checked
        explanation_parts.append("\nChecked for these entity types:")
        if self.selected_entities is not None:
            for entity in self.selected_entities:
                explanation_parts.append(f"- {entity}")
        return "\n".join(explanation_parts)

    @weave.op
    def anonymize_text(
        self,
        output: str,
        analyzer_results: list["RecognizerResult"],
        detected_entities: dict[str, list[str]],
    ) -> Optional[str]:
        anonymized_text = None
        if detected_entities and self._anonymizer is not None:
            anonymized_result = self._anonymizer.anonymize(
                text=output, analyzer_results=analyzer_results
            )
            anonymized_text = anonymized_result.text
        return anonymized_text

    @weave.op
    def score(
        self, output: str, entities: Optional[list[str]] = None
    ) -> WeaveScorerResult:
        if self._analyzer is None:
            raise ValueError("Analyzer is not initialized")
        if entities is None:
            entities = self.selected_entities
        analyzer_results = self._analyzer.analyze(
            text=str(output), entities=entities, language=self.language
        )
        detected_entities = self.group_analyzer_results_by_entity_type(
            output, analyzer_results
        )
        reason = self.create_reason(detected_entities)
        anonymized_text = self.anonymize_text(
            output, analyzer_results, detected_entities
        )
        return WeaveScorerResult(
            passed=not bool(detected_entities),
            metadata={
                "detected_entities": detected_entities,
                "reason": reason,
                "anonymized_text": anonymized_text,
            },
        )



from importlib import metadata

from packaging import version

try:
    mistral_version = metadata.version("mistralai")
except metadata.PackageNotFoundError:
    mistral_version = "1.0"  # we need to return a patching function

if version.parse(mistral_version) < version.parse("1.0.0"):
    from .v0.mistral import get_mistral_patcher  # noqa: F401

    print(
        f"Using MistralAI version {mistral_version}. Please consider upgrading to version 1.0.0 or later."
    )
else:
    from .v1.mistral import get_mistral_patcher  # noqa: F401



from __future__ import annotations

import importlib
import os
from typing import Any, Callable

from pydantic import AnyUrl

import weave
from weave.integrations.patcher import MultiPatcher, NoOpPatcher, SymbolPatcher
from weave.trace.autopatch import IntegrationSettings, OpSettings

_mcp_client_patcher: MultiPatcher | None = None


def mcp_client_wrapper(settings: OpSettings) -> Callable:
    """Wrapper for MCP client methods."""

    def wrapper(fn: Callable) -> Callable:
        op_kwargs = settings.model_dump()
        op_name = op_kwargs.get("name", "")

        # Special handlings to create synthetic child calls to surface more information
        if op_name.endswith("call_tool"):

            def wrapped_call_tool(*args: Any, **kwargs: Any) -> Any:
                # ClientSession.call_tool(self, name, arguments) is the expected signature
                tool_name = None

                # Try to get tool name from args or kwargs
                if len(args) >= 2:
                    tool_name = args[1]
                elif "name" in kwargs:
                    tool_name = kwargs["name"]

                if tool_name:
                    # Create a new settings object with tool name included in the name and display name
                    base_name = op_name or "mcp.client.session.ClientSession.call_tool"

                    # Set the op name to include the tool name
                    new_op_name = f"{base_name}.{tool_name}"

                    # Use just the tool name as the display name
                    new_settings = settings.model_copy(
                        update={"name": new_op_name, "call_display_name": tool_name}
                    )

                    # Create a new op with the updated settings
                    new_op_kwargs = new_settings.model_dump()
                    op = weave.op(fn, **new_op_kwargs)
                    return op(*args, **kwargs)

                # If we couldn't determine the tool name, just use the original op
                op = weave.op(fn, **op_kwargs)
                return op(*args, **kwargs)

            return wrapped_call_tool
        elif op_name.endswith("read_resource"):

            def wrapped_read_resource(*args: Any, **kwargs: Any) -> Any:
                display_uri = None

                # uri is an object of pydantic `AnyUrl`
                if len(args) >= 2:
                    uri = args[1]
                elif "uri" in kwargs:
                    uri = kwargs["uri"]

                if isinstance(uri, AnyUrl):
                    display_uri = f"{uri.scheme}://{uri.host}"

                if display_uri:
                    base_name = (
                        op_name or "mcp.client.session.ClientSession.read_resource"
                    )
                    new_op_name = f"{base_name}.{display_uri}"
                    new_settings = settings.model_copy(
                        update={"name": new_op_name, "call_display_name": display_uri}
                    )
                    new_op_kwargs = new_settings.model_dump()
                    op = weave.op(fn, **new_op_kwargs)
                    return op(*args, **kwargs)

                op = weave.op(fn, **op_kwargs)
                return op(*args, **kwargs)

            return wrapped_read_resource
        elif op_name.endswith("get_prompt"):

            def wrapped_get_prompt(*args: Any, **kwargs: Any) -> Any:
                prompt_name = None

                if len(args) >= 2:
                    prompt_name = args[1]
                elif "name" in kwargs:
                    prompt_name = kwargs["name"]

                if prompt_name:
                    base_name = op_name or "mcp.client.session.ClientSession.get_prompt"
                    new_op_name = f"{base_name}.{prompt_name}"
                    new_settings = settings.model_copy(
                        update={"name": new_op_name, "call_display_name": prompt_name}
                    )
                    new_op_kwargs = new_settings.model_dump()
                    op = weave.op(fn, **new_op_kwargs)
                    return op(*args, **kwargs)

                op = weave.op(fn, **op_kwargs)
                return op(*args, **kwargs)

            return wrapped_get_prompt
        else:
            # For other methods, use the standard wrapper
            op = weave.op(fn, **op_kwargs)
            return op

    return wrapper


def get_mcp_client_patcher(
    settings: IntegrationSettings | None = None,
) -> MultiPatcher | NoOpPatcher:
    if settings is None:
        settings = IntegrationSettings()

    if not settings.enabled:
        return NoOpPatcher()

    global _mcp_client_patcher
    if _mcp_client_patcher is not None:
        return _mcp_client_patcher

    base = settings.op_settings

    # Settings for core client methods
    call_tool_settings = base.model_copy(
        update={
            "name": base.name or "mcp.client.session.ClientSession.call_tool",
            "call_display_name": base.call_display_name or "ClientSession.call_tool",
        }
    )

    list_tools_settings = base.model_copy(
        update={
            "name": base.name or "mcp.client.session.ClientSession.list_tools",
            "call_display_name": base.call_display_name or "ClientSession.list_tools",
        }
    )

    read_resource_settings = base.model_copy(
        update={
            "name": base.name or "mcp.client.session.ClientSession.read_resource",
            "call_display_name": base.call_display_name
            or "ClientSession.read_resource",
        }
    )

    list_resources_settings = base.model_copy(
        update={
            "name": base.name or "mcp.client.session.ClientSession.list_resources",
            "call_display_name": base.call_display_name
            or "ClientSession.list_resources",
        }
    )

    list_prompts_settings = base.model_copy(
        update={
            "name": base.name or "mcp.client.session.ClientSession.list_prompts",
            "call_display_name": base.call_display_name or "ClientSession.list_prompts",
        }
    )

    get_prompt_settings = base.model_copy(
        update={
            "name": base.name or "mcp.client.session.ClientSession.get_prompt",
            "call_display_name": base.call_display_name or "ClientSession.get_prompt",
        }
    )

    # Create patchers for all methods we want to trace
    patchers = [
        # Core client methods
        SymbolPatcher(
            lambda: importlib.import_module("mcp.client.session"),
            "ClientSession.call_tool",
            mcp_client_wrapper(call_tool_settings),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("mcp.client.session"),
            "ClientSession.read_resource",
            mcp_client_wrapper(read_resource_settings),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("mcp.client.session"),
            "ClientSession.get_prompt",
            mcp_client_wrapper(get_prompt_settings),
        ),
    ]

    trace_list_operations = os.environ.get("MCP_TRACE_LIST_OPERATIONS", "").lower() in (
        "true",
        "1",
        "yes",
    )
    if trace_list_operations:
        patchers.extend(
            [
                SymbolPatcher(
                    lambda: importlib.import_module("mcp.client.session"),
                    "ClientSession.list_tools",
                    mcp_client_wrapper(list_tools_settings),
                ),
                SymbolPatcher(
                    lambda: importlib.import_module("mcp.client.session"),
                    "ClientSession.list_resources",
                    mcp_client_wrapper(list_resources_settings),
                ),
                SymbolPatcher(
                    lambda: importlib.import_module("mcp.client.session"),
                    "ClientSession.list_prompts",
                    mcp_client_wrapper(list_prompts_settings),
                ),
            ]
        )

    _mcp_client_patcher = MultiPatcher(patchers)

    return _mcp_client_patcher



from __future__ import annotations

import urllib
from concurrent.futures import Future
from dataclasses import asdict, dataclass, fields
from datetime import datetime
from typing import Any, Union, cast

from weave.trace_server import refs_internal
from weave.trace_server.errors import ObjectDeletedError

DICT_KEY_EDGE_NAME = refs_internal.DICT_KEY_EDGE_NAME
LIST_INDEX_EDGE_NAME = refs_internal.LIST_INDEX_EDGE_NAME
OBJECT_ATTR_EDGE_NAME = refs_internal.OBJECT_ATTR_EDGE_NAME
TABLE_ROW_ID_EDGE_NAME = refs_internal.TABLE_ROW_ID_EDGE_NAME


class WeaveDigestError(ValueError):
    """Raised when a digest is invalid."""


@dataclass(frozen=True)
class Ref:
    def uri(self) -> str:
        raise NotImplementedError

    def as_param_dict(self) -> dict:
        return asdict(self)

    def __deepcopy__(self, memo: dict) -> Ref:
        d = {f.name: getattr(self, f.name) for f in fields(self)}
        res = self.__class__(**d)
        memo[id(self)] = res
        return res


@dataclass(frozen=True)
class TableRef(Ref):
    entity: str
    project: str
    _digest: str | Future[str]
    _row_digests: list[str] | Future[list[str]] | None = None

    def as_param_dict(self) -> dict:
        return {
            "entity": self.entity,
            "project": self.project,
            "_digest": self._digest,
            "_row_digests": self._row_digests,
        }

    @property
    def digest(self) -> str:
        if isinstance(self._digest, Future):
            # Block until the Future resolves and store the result
            self.__dict__["_digest"] = self._digest.result()

        if not isinstance(self._digest, str):
            raise WeaveDigestError(f"TableRef digest is not a string: {self._digest}")

        refs_internal.validate_no_slashes(self._digest, "digest")
        refs_internal.validate_no_colons(self._digest, "digest")

        return self._digest

    @property
    def row_digests(self) -> list[str]:
        if isinstance(self._row_digests, Future):
            # Block until the Future resolves and store the result
            self.__dict__["_row_digests"] = self._row_digests.result()

        if not isinstance(self._row_digests, list):
            raise WeaveDigestError(
                f"TableRef row_digests is not a list: {self._row_digests}"
            )

        return self._row_digests

    @property
    def project_id(self) -> str:
        return f"{self.entity}/{self.project}"

    def __post_init__(self) -> None:
        if isinstance(self._digest, str):
            refs_internal.validate_no_slashes(self._digest, "digest")
            refs_internal.validate_no_colons(self._digest, "digest")

    def uri(self) -> str:
        return f"weave:///{self.entity}/{self.project}/table/{self.digest}"


@dataclass(frozen=True)
class RefWithExtra(Ref):
    def with_extra(self, extra: tuple[str | Future[str], ...]) -> RefWithExtra:
        params = self.as_param_dict()
        params["_extra"] = self._extra + tuple(extra)  # type: ignore
        return self.__class__(**params)

    def with_key(self, key: str) -> RefWithExtra:
        return self.with_extra((DICT_KEY_EDGE_NAME, key))

    def with_attr(self, attr: str) -> RefWithExtra:
        return self.with_extra((OBJECT_ATTR_EDGE_NAME, attr))

    def with_index(self, index: int) -> RefWithExtra:
        return self.with_extra((LIST_INDEX_EDGE_NAME, str(index)))

    def with_item(self, item_digest: str | Future[str]) -> RefWithExtra:
        return self.with_extra((TABLE_ROW_ID_EDGE_NAME, item_digest))


@dataclass(frozen=True)
class ObjectRef(RefWithExtra):
    entity: str
    project: str
    name: str
    _digest: str | Future[str]
    _extra: tuple[str | Future[str], ...] = ()

    def as_param_dict(self) -> dict:
        return {
            "entity": self.entity,
            "project": self.project,
            "name": self.name,
            "_digest": self._digest,
            "_extra": self._extra,
        }

    @property
    def extra(self) -> tuple[str, ...]:
        if any(isinstance(e, Future) for e in self._extra):
            self.__dict__["_extra"] = tuple(
                e if isinstance(e, str) else e.result() for e in self._extra
            )
            refs_internal.validate_extra(list(self.extra))

        return cast(tuple[str, ...], self._extra)

    @property
    def digest(self) -> str:
        if isinstance(self._digest, Future):
            # Block until the Future resolves and store the result
            self.__dict__["_digest"] = self._digest.result()

        if not isinstance(self._digest, str):
            raise WeaveDigestError(f"ObjectRef digest is not a string: {self._digest}")

        refs_internal.validate_no_slashes(self._digest, "digest")
        refs_internal.validate_no_colons(self._digest, "digest")

        return self._digest

    def __post_init__(self) -> None:
        if isinstance(self._digest, str):
            refs_internal.validate_no_slashes(self._digest, "digest")
            refs_internal.validate_no_colons(self._digest, "digest")

        refs_internal.validate_no_slashes(self.name, "name")
        refs_internal.validate_no_colons(self.name, "name")

    def uri(self) -> str:
        u = f"weave:///{self.entity}/{self.project}/object/{self.name}:{self.digest}"
        if self.extra:
            u += "/" + "/".join(refs_internal.extra_value_quoter(e) for e in self.extra)
        return u

    def get(self, *, objectify: bool = True) -> Any:
        # Move import here so that it only happens when the function is called.
        # This import is invalid in the trace server and represents a dependency
        # that should be removed.
        from weave.trace.context.weave_client_context import get_weave_client
        from weave.trace.weave_init import init_weave

        gc = get_weave_client()
        if gc is not None:
            return gc.get(self, objectify=objectify)

        # Special case: If the user is attempting to fetch an object but has not
        # yet initialized the client, we can initialize a client to
        # fetch the object. It is critical to reset the client after fetching the
        # object to avoid any side effects in user code.
        init_client = init_weave(
            f"{self.entity}/{self.project}", ensure_project_exists=False
        )
        try:
            res = init_client.client.get(self, objectify=objectify)
        finally:
            init_client.reset()
        return res

    def is_descended_from(self, potential_ancestor: ObjectRef) -> bool:
        if self.entity != potential_ancestor.entity:
            return False
        if self.project != potential_ancestor.project:
            return False
        if self.name != potential_ancestor.name:
            return False
        if self.digest != potential_ancestor.digest:
            return False
        if len(self.extra) <= len(potential_ancestor.extra):
            return False
        return all(
            self.extra[i] == potential_ancestor.extra[i]
            for i in range(len(potential_ancestor.extra))
        )

    def delete(self) -> None:
        from weave.trace.context.weave_client_context import get_weave_client

        gc = get_weave_client()
        if gc is not None:
            gc.delete_object_version(self)


@dataclass(frozen=True)
class OpRef(ObjectRef):
    def uri(self) -> str:
        u = f"weave:///{self.entity}/{self.project}/op/{self.name}:{self.digest}"
        if self.extra:
            u += "/" + "/".join(refs_internal.extra_value_quoter(e) for e in self.extra)
        return u

    def delete(self) -> None:
        from weave.trace.context.weave_client_context import get_weave_client

        gc = get_weave_client()
        if gc is not None:
            gc.delete_op_version(self)


@dataclass(frozen=True)
class CallRef(RefWithExtra):
    entity: str
    project: str
    id: str
    _extra: tuple[str | Future[str], ...] = ()

    def as_param_dict(self) -> dict:
        return {
            "entity": self.entity,
            "project": self.project,
            "id": self.id,
            "_extra": self._extra,
        }

    @property
    def extra(self) -> tuple[str, ...]:
        return tuple(e if isinstance(e, str) else e.result() for e in self._extra)

    def uri(self) -> str:
        u = f"weave:///{self.entity}/{self.project}/call/{self.id}"
        if self._extra:
            u += "/" + "/".join(refs_internal.extra_value_quoter(e) for e in self.extra)
        return u


@dataclass(frozen=True)
class DeletedRef(Ref):
    ref: Ref
    deleted_at: datetime
    error: ObjectDeletedError

    def __repr__(self) -> str:
        return f"<DeletedRef {self.uri()}>"

    def uri(self) -> str:
        return self.ref.uri()


AnyRef = Union[ObjectRef, TableRef, CallRef, OpRef]


def parse_name_version(name_version: str) -> tuple[str, str]:
    if ":" in name_version:
        name, version = name_version.rsplit(":", maxsplit=1)
        return name, version
    return name_version, "latest"


def parse_uri(uri: str) -> AnyRef:
    if not uri.startswith("weave:///"):
        raise ValueError(f"Invalid URI: {uri}")
    path = uri[len("weave:///") :]
    parts = path.split("/")
    if len(parts) < 3:
        raise ValueError(f"Invalid URI: {uri}")
    entity, project, kind = parts[:3]
    remaining = tuple(parts[3:])
    if kind == "table":
        return TableRef(entity=entity, project=project, _digest=remaining[0])
    extra = tuple(urllib.parse.unquote(r) for r in remaining[1:])
    if kind == "call":
        return CallRef(entity=entity, project=project, id=remaining[0], _extra=extra)
    elif kind == "object":
        name, version = parse_name_version(remaining[0])
        return ObjectRef(
            entity=entity, project=project, name=name, _digest=version, _extra=extra
        )
    elif kind == "op":
        name, version = parse_name_version(remaining[0])
        return OpRef(
            entity=entity, project=project, name=name, _digest=version, _extra=extra
        )
    else:
        raise ValueError(f"Unknown ref kind: {kind}")


def parse_op_uri(uri: str) -> OpRef:
    if not isinstance(parsed := parse_uri(uri), OpRef):
        raise TypeError(f"URI is not for an Op: {uri}")
    return parsed


def parse_object_uri(uri: str) -> ObjectRef:
    if not isinstance(parsed := parse_uri(uri), ObjectRef):
        raise TypeError(f"URI is not for an Object: {uri}")
    return parsed


def maybe_parse_uri(s: str) -> AnyRef | None:
    try:
        return parse_uri(s)
    except ValueError:
        return None



"""
This module contains the classes and utilities for working with file storage URIs.

A URI has specific form and structure and therefore we need to parse and validate them.
Each storage type (S3, GCS, Azure) has its own URI format and validation rules.
"""

from urllib.parse import ParseResult, urlparse


class URIParseError(ValueError):
    """Raised when a storage URI cannot be parsed or is invalid."""

    pass


class FileStorageURI:
    """Base class for all file storage URIs.

    This class provides the interface and common functionality for handling
    different types of storage URIs (S3, GCS, Azure). Each subclass implements
    specific parsing and validation rules for its storage type.

    Attributes:
        scheme (str): The URI scheme (e.g., 's3', 'gs', 'az')
        path (str): The path component of the URI
    """

    scheme: str
    path: str

    @classmethod
    def _from_parse_result(cls, parsed_uri: ParseResult) -> "FileStorageURI":
        """Create a storage URI instance from a parsed URI.

        Args:
            parsed_uri: A ParseResult instance from urlparse

        Returns:
            A new instance of the appropriate FileStorageURI subclass

        Raises:
            URIParseError: If the URI is invalid for the storage type
        """
        raise NotImplementedError("Subclasses must implement this method")

    def _to_parse_result(self) -> ParseResult:
        """Convert the storage URI to a ParseResult.

        Returns:
            ParseResult containing the components of the URI
        """
        raise NotImplementedError("Subclasses must implement this method")

    def has_path(self) -> bool:
        return self.path != ""

    def with_path(self, path: str) -> "FileStorageURI":
        copied = self.parse_uri_str(self.to_uri_str())
        copied.path = path
        return copied

    def to_uri_str(self) -> str:
        return self._to_parse_result().geturl()

    @classmethod
    def parse_uri_str(cls, uri: str) -> "FileStorageURI":
        parsed_uri = urlparse(uri)
        scheme = parsed_uri.scheme

        # Get all subclasses of FileStorageURI
        candidate_classes = cls.__subclasses__()
        if cls != FileStorageURI:
            candidate_classes.append(cls)
        for candidate_class in candidate_classes:
            if candidate_class.scheme == scheme:
                return candidate_class._from_parse_result(parsed_uri)
        raise URIParseError(f"No matching scheme for file storage URI: {uri}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.to_uri_str()})"


class S3FileStorageURI(FileStorageURI):
    """Amazon S3 storage URI handler.

    Format: s3://<bucket>/<path>
    Example: s3://my-bucket/path/to/object

    Attributes:
        scheme (str): Always 's3'
        bucket (str): The S3 bucket name
        path (str): The object key/path within the bucket
    """

    scheme = "s3"
    bucket: str

    def __init__(self, bucket: str, path: str):
        self.bucket = bucket.strip("/")
        if self.bucket == "":
            raise URIParseError("Bucket cannot be empty")
        self.path = path.strip("/")

    @classmethod
    def _from_parse_result(cls, parsed_uri: ParseResult) -> "S3FileStorageURI":
        if parsed_uri.scheme != cls.scheme:
            raise URIParseError(
                f"Incorrect scheme for S3 file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.params != "":
            raise URIParseError(
                f"Invalid params for S3 file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.query != "":
            raise URIParseError(
                f"Invalid query for S3 file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.fragment != "":
            raise URIParseError(
                f"Invalid fragment for S3 file storage URI: {parsed_uri.geturl()}"
            )
        return cls(parsed_uri.netloc, parsed_uri.path)

    def _to_parse_result(self) -> ParseResult:
        return ParseResult(
            scheme=self.scheme,
            netloc=self.bucket,
            path=self.path,
            params="",
            query="",
            fragment="",
        )


class GCSFileStorageURI(FileStorageURI):
    """Google Cloud Storage URI handler.

    Format: gs://<bucket>/<path>
    Example: gs://my-bucket/path/to/object

    Attributes:
        scheme (str): Always 'gs'
        bucket (str): The GCS bucket name
        path (str): The object path within the bucket
    """

    scheme = "gs"
    bucket: str

    def __init__(self, bucket: str, path: str):
        self.bucket = bucket.strip("/")
        if self.bucket == "":
            raise URIParseError("Bucket cannot be empty")
        self.path = path.strip("/")

    @classmethod
    def _from_parse_result(cls, parsed_uri: ParseResult) -> "GCSFileStorageURI":
        if parsed_uri.scheme != cls.scheme:
            raise URIParseError(
                f"Incorrect scheme for GCS file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.params != "":
            raise URIParseError(
                f"Invalid params for GCS file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.query != "":
            raise URIParseError(
                f"Invalid query for GCS file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.fragment != "":
            raise URIParseError(
                f"Invalid fragment for GCS file storage URI: {parsed_uri.geturl()}"
            )
        return cls(parsed_uri.netloc, parsed_uri.path)

    def _to_parse_result(self) -> ParseResult:
        return ParseResult(
            scheme=self.scheme,
            netloc=self.bucket,
            path=self.path,
            params="",
            query="",
            fragment="",
        )


class AzureFileStorageURI(FileStorageURI):
    """Azure Blob Storage URI handler.

    Format: az://<account>/<container>/<path>
    Example: az://myaccount/mycontainer/path/to/blob

    The account is stored in the netloc component, while container and path
    are stored in the path component of the URI.
    """

    scheme = "az"
    account: str
    container: str

    def __init__(self, account: str, container: str, path: str):
        self.account = account.strip("/")
        if self.account == "":
            raise URIParseError("Account cannot be empty")
        self.container = container.strip("/")
        if self.container == "":
            raise URIParseError("Container cannot be empty")
        self.path = path.strip("/")

    @classmethod
    def _from_parse_result(cls, parsed_uri: ParseResult) -> "AzureFileStorageURI":
        if parsed_uri.scheme != cls.scheme:
            raise URIParseError(
                f"Incorrect scheme for Azure file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.params != "":
            raise URIParseError(
                f"Invalid params for Azure file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.query != "":
            raise URIParseError(
                f"Invalid query for Azure file storage URI: {parsed_uri.geturl()}"
            )
        if parsed_uri.fragment != "":
            raise URIParseError(
                f"Invalid fragment for Azure file storage URI: {parsed_uri.geturl()}"
            )

        # Split the path into container and path
        path = parsed_uri.path.strip("/")
        if path == "":
            raise URIParseError(
                f"Invalid path for Azure file storage URI - must include container name: {parsed_uri.geturl()}"
            )
        path_parts = path.split("/", 1)
        account = parsed_uri.netloc
        container = path_parts[0]
        path = path_parts[1] if len(path_parts) > 1 else ""
        return cls(account, container, path)

    def _to_parse_result(self) -> ParseResult:
        path = self.container
        if self.path != "":
            path += "/" + self.path
        return ParseResult(
            scheme=self.scheme,
            netloc=self.account,
            path=path,
            params="",
            query="",
            fragment="",
        )



from __future__ import annotations

import configparser
import logging
import netrc
import os
from urllib.parse import urlparse

WEAVE_PARALLELISM = "WEAVE_PARALLELISM"

logger = logging.getLogger(__name__)


class Settings:
    """A minimal readonly implementation of wandb/old/settings.py for reading settings"""

    DEFAULT_SECTION = "default"
    DEFAULT_BASE_URL = "https://api.wandb.ai"

    def __init__(self) -> None:
        self._settings = configparser.ConfigParser()
        if not self._settings.has_section(self.DEFAULT_SECTION):
            self._settings.add_section(self.DEFAULT_SECTION)
        self._settings.read([Settings._global_path(), Settings._local_path()])

    @property
    def base_url(self) -> str:
        try:
            return self._settings.get(self.DEFAULT_SECTION, "base_url")
        except configparser.NoOptionError:
            return self.DEFAULT_BASE_URL

    @staticmethod
    def _local_path() -> str:
        return os.path.join(os.getcwd(), "wandb", "settings")

    @staticmethod
    def _global_path() -> str:
        default_config_dir = os.path.join(os.path.expanduser("~"), ".config", "wandb")
        config_dir = os.environ.get("WANDB_CONFIG_DIR", default_config_dir)
        return os.path.join(config_dir, "settings")


def get_weave_parallelism() -> int:
    return int(os.getenv(WEAVE_PARALLELISM, "20"))


def wandb_base_url() -> str:
    settings = Settings()
    return os.environ.get("WANDB_BASE_URL", settings.base_url).rstrip("/")


def wandb_frontend_base_url() -> str:
    public_url = os.getenv("WANDB_PUBLIC_BASE_URL", "").rstrip("/")
    return public_url if public_url != "" else wandb_base_url()


def weave_trace_server_url() -> str:
    base_url = wandb_frontend_base_url()
    default = "https://trace.wandb.ai"
    if base_url != "https://api.wandb.ai":
        default = base_url + "/traces"
    return os.getenv("WF_TRACE_SERVER_URL", default)


def _wandb_api_key_via_env() -> str | None:
    api_key = os.environ.get("WANDB_API_KEY")
    return api_key


def _wandb_api_key_via_netrc() -> str | None:
    for filepath in ("~/.netrc", "~/_netrc"):
        api_key = _wandb_api_key_via_netrc_file(filepath)
        if api_key:
            return api_key
    return None


def _wandb_api_key_via_netrc_file(filepath: str) -> str | None:
    netrc_path = os.path.expanduser(filepath)
    if not os.path.exists(netrc_path):
        return None
    nrc = netrc.netrc(netrc_path)
    res = nrc.authenticators(urlparse(wandb_base_url()).netloc)
    api_key = None
    if res:
        _, _, api_key = res
    return api_key


def weave_wandb_api_key() -> str | None:
    env_api_key = _wandb_api_key_via_env()
    netrc_api_key = _wandb_api_key_via_netrc()
    if env_api_key and netrc_api_key and env_api_key != netrc_api_key:
        logger.warning(
            "There are different credentials in the netrc file and the environment. Using the environment value."
        )
    return env_api_key or netrc_api_key



"""
Technical Note:
This LangChain integration patching differs from other integrations in how tracing is enabled:

1. Environment Variable:
   Unlike other integrations where tracing is typically enabled through the `SymbolPatcher`,
   LangChain tracing is enabled by setting an environment variable (WEAVE_TRACE_LANGCHAIN).
   This is because LangChain configures tracing during runtime, not at import time.

2. Runtime Configuration:
   The `LangchainPatcher` class handles setting up the tracing hook based on this environment variable.
   It uses `register_configure_hook` to set up the tracing callback at runtime:

   register_configure_hook(
       weave_tracing_callback_var, True, WeaveTracer, "WEAVE_TRACE_LANGCHAIN"
   )

3. Respecting User Settings:
   The patcher respects any existing WEAVE_TRACE_LANGCHAIN environment variable set by the user:
   - If not set, it's set to "true" and global patching is enabled.
   - If already set, its value is preserved

4. Context Manager:
   Tracing can be enabled in code using the `weave_tracing_enabled()` context manager:

   with weave_tracing_enabled():
       # LangChain code here

This approach allows for more flexible runtime configuration while still respecting global user-defined preferences.
"""

import datetime
import json
from contextlib import contextmanager
from contextvars import ContextVar
from uuid import UUID

from weave.integrations.integration_utilities import (
    make_pythonic_function_name,
    truncate_op_name,
)
from weave.integrations.patcher import Patcher
from weave.trace.context import call_context
from weave.trace.context import weave_client_context as weave_client_context
from weave.trace.weave_client import Call

import_failed = False

try:
    from langchain_core.load import dumpd
    from langchain_core.messages import BaseMessage
    from langchain_core.tracers import Run
    from langchain_core.tracers.base import BaseTracer
    from langchain_core.tracers.context import register_configure_hook
except ImportError:
    import_failed = True

from collections.abc import Generator
from typing import Any, Optional

RUNNABLE_SEQUENCE_NAME = "RunnableSequence"

if not import_failed:

    def _run_to_dict(run: Run, as_input: bool = False) -> dict:
        run_dict = run.json(
            exclude={
                "child_runs",
                "inputs",
                "outputs",
                "serialized",
                "events",
                "reference_example_id",
                "trace_id",
                "dotted_order",
            },
            exclude_unset=True,
            exclude_none=True,
            exclude_defaults=True,
        )

        run_dict = json.loads(run_dict)
        if as_input:
            run_dict["inputs"] = run.inputs.copy() if run.inputs is not None else None
        else:
            run_dict["outputs"] = (
                run.outputs.copy() if run.outputs is not None else None,
            )

        run_dict = {k: v for k, v in run_dict.items() if v}
        return run_dict

    class WeaveTracer(BaseTracer):  # pyright: ignore[reportRedeclaration]
        run_inline: bool = True

        def __init__(self, **kwargs: Any) -> None:
            self._call_map: dict[str, Call] = {}
            self.latest_run: Optional[Run] = None
            self.gc = weave_client_context.require_weave_client()
            super().__init__()

        def _persist_run(self, run: Run) -> None:
            run_ = run.copy()
            self.latest_run = run_

        def _persist_run_single(self, run: Run) -> None:
            """Persist a run."""
            run_dict = _run_to_dict(run, as_input=True)

            """Now we must determine the parent_run to associate this call with.
            In most cases, it is very straight forward - we just look up the
            parent_run_id in the call map and associate it with the parent call.

            However, there is a very specific case with `RunnableSequence`. The way that
            `RunnableSequence::batch` is implemented, Langchain will fire off two runs
            in sequence. For example, in the event that we have a model like:
            ```
            (prompt | llm).batch([1, 2])
            ```
            Langchain's sequence of call starts is:
            ```
            start RunnableSequence1 -> no parent
            start RunnableSequence2 -> no parent
            (2 parallel threads) to resolve prompt batch
            Thread 1:
                start Prompt1 -> RunnableSequence1
                end   Prompt1
            Thread 2:
                start Prompt2 -> RunnableSequence2
                end   Prompt1
            (2 parallel threads)  to resolve llm batch
            Thread 1:
                start LLM1 -> RunnableSequence1
                (OpenAI calls run in here)
                end   LLM1
            Thread 2:
                start LLM2 -> RunnableSequence2
                (OpenAI calls run in here)
                end   LLM2
            end RunnableSequence1
            end RunnableSequence2
            ```

            In these cases, RunnableSequence2 is started, but our stack context will have
            `RunnableSequence1` popped onto the stack. To solve for this, we need to send
            `False` as the `parent_id`, telling the system: "trust me, this is a root".
            """
            parent_run: Optional[Call] = None
            lc_parent_run_id = (
                str(run.parent_run_id) if run.parent_run_id is not None else None
            )
            wv_parent_run = (
                self._call_map.get(lc_parent_run_id) if lc_parent_run_id else None
            )
            use_stack = True
            if wv_parent_run is not None:
                parent_run = wv_parent_run
            else:
                # Here is our check for the specific condition
                wv_current_run = call_context.get_current_call()

                # First, there needs to be something on the stack.
                if wv_current_run is not None:
                    attrs = call_context.call_attributes.get()
                    # Now, the major condition:
                    if (
                        # 1. Both runs must be of type `RunnableSequence`
                        attrs.get("lc_name") == run.name == RUNNABLE_SEQUENCE_NAME
                        # 2. Both us and the sibling run must have empty parents (i think
                        # this condition will always be true, else we would have a parent
                        # run already, but trying to be safe here)
                        and attrs.get("parent_run_id") == lc_parent_run_id == None
                    ):
                        # Now, we know that Langchain has confused us. And we want to set the
                        # parent to the current Weave Run's parent. So, if that parent is
                        # None, then we use `False` here to force a root. Else we lookup
                        # the parent from the client. You might be thinking... id the parent
                        # run_id is none, when would this NOT be none? The answer: when Langchain
                        # is called inside of another op (like Evaluations!)
                        if wv_current_run.parent_id is None:
                            use_stack = False
                        else:
                            # Hack in memory parent call to satisfy `create_call` without actually
                            # getting the parent.
                            parent_run = Call(
                                id=wv_current_run.parent_id,
                                trace_id=wv_current_run.trace_id,
                                _op_name="",
                                project_id="",
                                parent_id=None,
                                inputs={},
                            )

            fn_name = make_pythonic_function_name(run.name)
            complete_op_name = f"langchain.{run.run_type.capitalize()}.{fn_name}"
            complete_op_name = truncate_op_name(complete_op_name)
            call_attrs = call_context.call_attributes.get()
            call_attrs.update(
                {
                    "lc_id": str(run.id),
                    "parent_run_id": lc_parent_run_id,
                    "lc_name": run.name,
                }
            )
            call = self.gc.create_call(
                # Make sure to add the run name once the UI issue is figured out
                complete_op_name,
                inputs=run_dict.get("inputs", {}),
                parent=parent_run,
                attributes=call_attrs,
                display_name=f"langchain.{run.run_type.capitalize()}.{run.name}",
                use_stack=use_stack,
            )

            # Add the call to the call map.
            self._call_map[str(run.id)] = call

        def _finish_run(self, run: Run) -> None:
            """Finish a run."""
            # If the event is in the call map, finish the call.
            run_id = str(run.id)
            if run_id in self._call_map:
                # Finish the call.
                call = self._call_map.pop(run_id)
                run_dict = _run_to_dict(run, as_input=False)
                self.gc.finish_call(call, run_dict)

        def _update_run_error(self, run: Run) -> None:
            call = self._call_map.pop(str(run.id))
            if call:
                self.gc.finish_call(
                    call, _run_to_dict(run), exception=Exception(run.error)
                )

        def on_chat_model_start(
            self,
            serialized: dict[str, Any],
            messages: list[list[BaseMessage]],
            *,
            run_id: UUID,
            tags: Optional[list[str]] = None,
            parent_run_id: Optional[UUID] = None,
            metadata: Optional[dict[str, Any]] = None,
            name: Optional[str] = None,
            **kwargs: Any,
        ) -> Run:
            """Start a trace for an LLM run."""
            start_time = datetime.datetime.now(datetime.timezone.utc)
            if metadata:
                kwargs.update({"metadata": metadata})
            chat_model_run = Run(
                id=run_id,
                parent_run_id=parent_run_id,
                serialized=serialized,
                inputs={
                    "messages": [[dumpd(msg) for msg in batch] for batch in messages]
                },
                extra=kwargs,
                events=[{"name": "start", "time": start_time}],
                start_time=start_time,
                run_type="llm",
                tags=tags,
                name=name,  # type: ignore[arg-type]
            )
            self._start_trace(chat_model_run)
            self._on_chat_model_start(chat_model_run)
            return chat_model_run

        def _on_llm_start(self, run: Run) -> None:
            self._persist_run_single(run)

        def _on_llm_end(self, run: Run) -> None:
            self._finish_run(run)

        def _on_llm_error(self, run: Run) -> None:
            self._update_run_error(run)

        def _on_chat_model_start(self, run: Run) -> None:
            self._persist_run_single(run)

        def _on_chat_model_end(self, run: Run) -> None:
            self._finish_run(run)

        def _on_chat_model_error(self, run: Run) -> None:
            self._update_run_error(run)

        def _on_chain_start(self, run: Run) -> None:
            self._persist_run_single(run)

        def _on_chain_end(self, run: Run) -> None:
            self._finish_run(run)

        def _on_chain_error(self, run: Run) -> None:
            self._update_run_error(run)

        def _on_tool_start(self, run: Run) -> None:
            self._persist_run_single(run)

        def _on_tool_end(self, run: Run) -> None:
            self._finish_run(run)

        def _on_tool_error(self, run: Run) -> None:
            self._update_run_error(run)

        def _on_retriever_start(self, run: Run) -> None:
            self._persist_run_single(run)

        def _on_retriever_end(self, run: Run) -> None:
            self._finish_run(run)

        def _on_retriever_error(self, run: Run) -> None:
            self._update_run_error(run)

else:

    class WeaveTracer:  # type: ignore
        pass


weave_tracing_callback_var: ContextVar[Optional[WeaveTracer]] = ContextVar(
    "tracing_weave_callback", default=None
)


@contextmanager
def weave_tracing_enabled(
    session_name: str = "default",
) -> Generator[None, None, None]:
    """Get the WandbTracer in a context manager.

    Args:
        session_name (str, optional): The name of the session.
            Defaults to "default".

    Returns:
        None

    Example:
        >>> with weave_tracing_enabled() as session:
        ...     # Use the WeaveTracer session
    """
    cb = WeaveTracer()
    weave_tracing_callback_var.set(cb)
    yield None
    weave_tracing_callback_var.set(None)


class LangchainPatcher(Patcher):
    def __init__(self) -> None:
        pass

    def attempt_patch(self) -> bool:
        if import_failed:
            return False
        try:
            import os

            self.original_trace_state = os.environ.get("WEAVE_TRACE_LANGCHAIN")
            if self.original_trace_state is None:
                os.environ["WEAVE_TRACE_LANGCHAIN"] = "true"
            else:
                os.environ["WEAVE_TRACE_LANGCHAIN"] = self.original_trace_state
            register_configure_hook(
                weave_tracing_callback_var, True, WeaveTracer, "WEAVE_TRACE_LANGCHAIN"
            )
        except Exception:
            return False
        else:
            return True

    def undo_patch(self) -> bool:
        if not hasattr(self, "original_trace_state"):
            return False
        try:
            import os

            if self.original_trace_state is None:
                del os.environ["WEAVE_TRACE_LANGCHAIN"]
            else:
                os.environ["WEAVE_TRACE_LANGCHAIN"] = self.original_trace_state
            weave_tracing_callback_var.set(None)
        except Exception:
            return False
        else:
            return True


langchain_patcher = LangchainPatcher()



import asyncio
import logging
import multiprocessing
import random
from collections import defaultdict
from collections.abc import AsyncIterator, Awaitable, Iterable
from typing import Any, Callable, TypeVar

T = TypeVar("T")
U = TypeVar("U")

_shown_warnings = set()


def transpose(rows: list[dict]) -> dict[str, list]:
    cols = defaultdict(list)
    for row in rows:
        for k, v in row.items():
            cols[k].append(v)
    return dict(cols)


async def async_foreach(
    sequence: Iterable[T],
    func: Callable[[T], Awaitable[U]],
    max_concurrent_tasks: int,
) -> AsyncIterator[tuple[T, U]]:
    """Process items from a sequence concurrently with a maximum number of parallel tasks.

    This function loads items from the input sequence lazily to support large or infinite
    sequences. Items are processed and yielded in the same order as the input sequence.

    Args:
        sequence: An iterable of items to process. Items are loaded lazily.
        func: An async function that processes each item from the sequence.
        max_concurrent_tasks: Maximum number of items to process concurrently.

    Yields:
        Tuples of (original_item, processed_result) in the same order as the input sequence.

    Example:
        ```python
        async def process(x: int) -> str:
            await asyncio.sleep(1)  # Simulate async work
            return str(x * 2)

        async for item, result in async_foreach(range(10), process, max_concurrent_tasks=3):
            print(f"Processed {item} -> {result}")
        ```

    Notes:
        - If func raises an exception, it will be propagated to the caller
        - Memory usage is bounded by max_concurrent_tasks
        - All pending tasks are properly cleaned up on error or cancellation
        - Results are yielded in the same order as the input sequence
    """
    semaphore = asyncio.Semaphore(max_concurrent_tasks)
    active_tasks: list[asyncio.Task] = []

    async def process_item(item: T) -> tuple[T, U]:
        """Process a single item using the provided function with semaphore control."""
        async with semaphore:
            result = await func(item)
            return item, result

    def maybe_queue_next_task() -> None:
        """Attempt to queue the next task from the iterator if available."""
        try:
            item = next(iterator)
            task = asyncio.create_task(process_item(item))
            active_tasks.append(task)
        except StopIteration:
            pass

    iterator = iter(sequence)

    try:
        # Prime the initial set of tasks
        for _ in range(max_concurrent_tasks):
            maybe_queue_next_task()

        while active_tasks:
            # Always wait for the first task in the list to complete
            # This ensures we yield results in order
            task = active_tasks.pop(0)  # Remove completed task from front of list
            try:
                item, result = await task
                yield item, result

                # Add a new task if there are more items
                maybe_queue_next_task()
            except Exception:
                # Clean up remaining tasks before re-raising
                for t in active_tasks:
                    t.cancel()
                await asyncio.gather(*active_tasks, return_exceptions=True)
                raise

    except asyncio.CancelledError:
        # Clean up tasks if the caller cancels this coroutine
        for task in active_tasks:
            task.cancel()
        await asyncio.gather(*active_tasks, return_exceptions=True)
        raise


def _subproc(
    queue: multiprocessing.Queue, func: Callable, *args: Any, **kwargs: Any
) -> None:
    result = func(*args, **kwargs)
    queue.put(result)


def _run_in_process(
    func: Callable, args: tuple = (), kwargs: dict = {}
) -> tuple[multiprocessing.Process, multiprocessing.Queue]:
    """Run a function in a separate process and return the process object and a multiprocessing.Queue for the result."""
    queue: multiprocessing.Queue = multiprocessing.Queue()
    process: multiprocessing.Process = multiprocessing.Process(
        target=_subproc, args=(queue, func) + args, kwargs=kwargs
    )
    process.start()
    return process, queue


async def run_in_process_with_timeout(
    timeout: float, func: Callable, *args: Any, **kwargs: Any
) -> Any:
    """Run a function in a separate process with a timeout. Terminate the process if it exceeds the timeout."""
    # Note, on osx, multiprocessing uses spawn by default. With span, subprocesses will re-import main,
    # and incur the cost of bootstrapping python and all imports, which for weave is currently about 1.5s.
    # Scripts that use this can call 'multiprocessing.set_start_method("fork")' which is much faster (0.05s)
    # depending on their use case (there are some issues with fork on osx).
    loop = asyncio.get_running_loop()
    process, queue = _run_in_process(func, args, kwargs)

    try:
        # Wait for the process to complete or timeout
        await asyncio.wait_for(loop.run_in_executor(None, process.join), timeout)
    except asyncio.TimeoutError:
        print("Function execution exceeded the timeout. Terminating process.")
        process.terminate()
        process.join()  # Ensure process resources are cleaned up
        raise

    if process.exitcode == 0:
        return queue.get()  # Retrieve result from the queue
    else:
        raise ValueError(
            "Unhandled exception in subprocess. Exitcode: " + str(process.exitcode)
        )


def warn_once(logger: logging.Logger, message: str) -> None:
    """Display a warning message only once. If the message has already been shown, do nothing."""
    if message not in _shown_warnings:
        logger.warning(message)
        _shown_warnings.add(message)


def make_memorable_name() -> str:
    adjectives = [
        "brave",
        "bright",
        "calm",
        "charming",
        "clever",
        "daring",
        "dazzling",
        "eager",
        "elegant",
        "eloquent",
        "fierce",
        "friendly",
        "gentle",
        "graceful",
        "happy",
        "honest",
        "imaginative",
        "innocent",
        "joyful",
        "jubilant",
        "keen",
        "kind",
        "lively",
        "loyal",
        "merry",
        "nice",
        "noble",
        "optimistic",
        "proud",
        "quiet",
        "rich",
        "sweet",
        "tender",
        "unique",
        "wise",
        "zealous",
    ]

    nouns = [
        "bear",
        "bird",
        "breeze",
        "cedar",
        "cloud",
        "daisy",
        "dawn",
        "dolphin",
        "dusk",
        "eagle",
        "fish",
        "flower",
        "forest",
        "hill",
        "horizon",
        "island",
        "lake",
        "lion",
        "maple",
        "meadow",
        "moon",
        "mountain",
        "oak",
        "ocean",
        "pine",
        "plateau",
        "rain",
        "river",
        "rose",
        "star",
        "stream",
        "sun",
        "tiger",
        "tree",
        "valley",
        "whale",
        "wind",
        "wolf",
    ]

    adj = random.choice(adjectives)
    noun = random.choice(nouns)
    return f"{adj}-{noun}"


def short_str(obj: Any, limit: int = 25) -> str:
    str_val = str(obj)
    if len(str_val) > limit:
        return str_val[:limit] + "..."
    return str_val



from __future__ import annotations

from collections.abc import Sequence

from weave.integrations.dspy.dspy_utils import get_symbol_patcher
from weave.integrations.patcher import MultiPatcher, NoOpPatcher, Patcher
from weave.trace.autopatch import IntegrationSettings

_dspy_patcher: MultiPatcher | None = None


class DSPyPatcher(MultiPatcher):
    def __init__(self, patchers: Sequence[Patcher]) -> None:
        super().__init__(patchers)
        try:
            import dspy

            from weave.integrations.dspy.dspy_callback import WeaveCallback

            is_callback_present = False
            for callback in dspy.settings.callbacks:
                if isinstance(callback, WeaveCallback):
                    is_callback_present = True
                    break

            if not is_callback_present:
                dspy.settings.callbacks.append(WeaveCallback())
        except ImportError:
            pass


def get_dspy_patcher(
    settings: IntegrationSettings | None = None,
) -> MultiPatcher | NoOpPatcher:
    if settings is None:
        settings = IntegrationSettings()

    if not settings.enabled:
        return NoOpPatcher()

    global _dspy_patcher
    if _dspy_patcher is not None:
        return _dspy_patcher

    base = settings.op_settings

    _dspy_patcher = DSPyPatcher(
        [
            get_symbol_patcher("dspy", "LM.__call__", base),
            get_symbol_patcher("dspy", "Embedder.__call__", base),
            get_symbol_patcher("dspy", "ColBERTv2.__call__", base),
            get_symbol_patcher("dspy", "BootstrapFinetune.compile", base),
            get_symbol_patcher("dspy", "MIPROv2.compile", base),
            get_symbol_patcher("dspy", "LabeledFewShot.compile", base),
            get_symbol_patcher("dspy", "KNNFewShot.compile", base),
            get_symbol_patcher("dspy", "KNN.__call__", base),
            get_symbol_patcher("dspy", "Ensemble.compile", base),
            get_symbol_patcher("dspy", "COPRO.compile", base),
            get_symbol_patcher(
                "dspy", "BootstrapFewShotWithRandomSearch.compile", base
            ),
            get_symbol_patcher("dspy", "BootstrapFewShot.compile", base),
            get_symbol_patcher("dspy", "BetterTogether.compile", base),
            get_symbol_patcher("dspy", "Evaluate.__call__", base),
            get_symbol_patcher("dspy.retrievers", "Embeddings.__call__", base),
            get_symbol_patcher("dspy.retrievers", "Embeddings.forward", base),
            get_symbol_patcher("dspy", "PythonInterpreter.__call__", base),
            get_symbol_patcher("dspy", "PythonInterpreter.execute", base),
            get_symbol_patcher("dspy", "ChatAdapter.__call__", base),
            get_symbol_patcher("dspy", "JSONAdapter.__call__", base),
        ]
    )

    return _dspy_patcher



# Sqlite Trace Server

import contextvars
import datetime
import hashlib
import json
import sqlite3
import threading
from collections.abc import Iterator
from typing import Any, Optional, cast
from zoneinfo import ZoneInfo

import emoji
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (
    ExportTraceServiceRequest,
)

from weave.trace_server import refs_internal as ri
from weave.trace_server import trace_server_interface as tsi
from weave.trace_server.emoji_util import detone_emojis
from weave.trace_server.errors import (
    InvalidRequest,
    NotFoundError,
    ObjectDeletedError,
)
from weave.trace_server.feedback import (
    TABLE_FEEDBACK,
    validate_feedback_create_req,
    validate_feedback_purge_req,
)
from weave.trace_server.ids import generate_id
from weave.trace_server.interface import query as tsi_query
from weave.trace_server.object_class_util import process_incoming_object_val
from weave.trace_server.opentelemetry.python_spans import ResourceSpans
from weave.trace_server.orm import Row, quote_json_path
from weave.trace_server.trace_server_common import (
    assert_parameter_length_less_than_max,
    digest_is_version_like,
    empty_str_to_none,
    get_nested_key,
    hydrate_calls_with_feedback,
    make_derived_summary_fields,
    make_feedback_query_req,
    set_nested_key,
)
from weave.trace_server.trace_server_interface_util import (
    WILDCARD_ARTIFACT_VERSION_AND_PATH,
    assert_non_null_wb_user_id,
    bytes_digest,
    extract_refs_from_values,
    str_digest,
)
from weave.trace_server.validation import object_id_validator

_conn_cursor: contextvars.ContextVar[
    Optional[tuple[sqlite3.Connection, sqlite3.Cursor]]
] = contextvars.ContextVar("conn_cursor", default=None)


def get_conn_cursor(db_path: str) -> tuple[sqlite3.Connection, sqlite3.Cursor]:
    # conn_cursor = _conn_cursor.get()
    conn_cursor = None
    if conn_cursor is None:
        conn = sqlite3.connect(db_path)
        # Create an array reverse function.
        conn.create_function("reverse", 1, lambda x: x[::-1])
        cursor = conn.cursor()
        conn_cursor = (conn, cursor)
        _conn_cursor.set(conn_cursor)
    return conn_cursor


class SqliteTraceServer(tsi.TraceServerInterface):
    def __init__(self, db_path: str):
        self.lock = threading.Lock()
        self.db_path = db_path

    def drop_tables(self) -> None:
        conn, cursor = get_conn_cursor(self.db_path)
        cursor.execute(TABLE_FEEDBACK.drop_sql())
        cursor.execute("DROP TABLE IF EXISTS calls")
        cursor.execute("DROP TABLE IF EXISTS objects")
        cursor.execute("DROP TABLE IF EXISTS tables")
        cursor.execute("DROP TABLE IF EXISTS table_rows")

    def setup_tables(self) -> None:
        conn, cursor = get_conn_cursor(self.db_path)
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS calls (
                project_id TEXT,
                id TEXT PRIMARY KEY,
                trace_id TEXT,
                parent_id TEXT,
                op_name TEXT,
                started_at TEXT,
                ended_at TEXT,
                exception TEXT,
                attributes TEXT,
                inputs TEXT,
                input_refs TEXT,
                output TEXT,
                output_refs TEXT,
                summary TEXT,
                wb_user_id TEXT,
                wb_run_id TEXT,
                deleted_at TEXT,
                display_name TEXT
            )
        """
        )
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS objects (
                project_id TEXT,
                object_id TEXT,
                wb_user_id TEXT,
                created_at TEXT,
                kind TEXT,
                base_object_class TEXT,
                refs TEXT,
                val_dump TEXT,
                digest TEXT,
                version_index INTEGER,
                is_latest INTEGER,
                deleted_at TEXT,
                primary key (project_id, kind, object_id, digest)
            )
        """
        )
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS tables (
                project_id TEXT,
                digest TEXT UNIQUE,
                row_digests STRING
            )
            """
        )
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS table_rows (
                project_id TEXT,
                digest TEXT UNIQUE,
                val TEXT
            )
            """
        )
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS files (
                project_id TEXT,
                digest TEXT,
                val BLOB,
                primary key (project_id, digest)
            )
            """
        )
        cursor.execute(TABLE_FEEDBACK.create_sql())

    def call_start_batch(self, req: tsi.CallCreateBatchReq) -> tsi.CallCreateBatchRes:
        res = []
        for item in req.batch:
            if item.mode == "start":
                res.append(self.call_start(item.req))
            elif item.mode == "end":
                res.append(self.call_end(item.req))
            else:
                raise ValueError("Invalid mode")
        return tsi.CallCreateBatchRes(res=res)

    # Creates a new call
    def call_start(self, req: tsi.CallStartReq) -> tsi.CallStartRes:
        conn, cursor = get_conn_cursor(self.db_path)
        if req.start.trace_id is None:
            raise ValueError("trace_id is required")
        if req.start.id is None:
            raise ValueError("id is required")
        with self.lock:
            # Converts the user-provided call details into a clickhouse schema.
            # This does validation and conversion of the input data as well
            # as enforcing business rules and defaults
            cursor.execute(
                """INSERT INTO calls (
                    project_id,
                    id,
                    trace_id,
                    parent_id,
                    op_name,
                    display_name,
                    started_at,
                    attributes,
                    inputs,
                    input_refs,
                    wb_user_id,
                    wb_run_id
                ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
                (
                    req.start.project_id,
                    req.start.id,
                    req.start.trace_id,
                    req.start.parent_id,
                    req.start.op_name,
                    req.start.display_name,
                    req.start.started_at.isoformat(),
                    json.dumps(req.start.attributes),
                    json.dumps(req.start.inputs),
                    json.dumps(
                        extract_refs_from_values(list(req.start.inputs.values()))
                    ),
                    req.start.wb_user_id,
                    req.start.wb_run_id,
                ),
            )
            conn.commit()

        # Returns the id of the newly created call
        return tsi.CallStartRes(
            id=req.start.id,
            trace_id=req.start.trace_id,
        )

    def call_end(self, req: tsi.CallEndReq) -> tsi.CallEndRes:
        conn, cursor = get_conn_cursor(self.db_path)
        parsable_output = req.end.output
        if not isinstance(parsable_output, dict):
            parsable_output = {"output": parsable_output}
        parsable_output = cast(dict, parsable_output)
        with self.lock:
            cursor.execute(
                """UPDATE calls SET
                    ended_at = ?,
                    exception = ?,
                    output = ?,
                    output_refs = ?,
                    summary = ?
                WHERE id = ?""",
                (
                    req.end.ended_at.isoformat(),
                    req.end.exception,
                    json.dumps(req.end.output),
                    json.dumps(
                        extract_refs_from_values(list(parsable_output.values()))
                    ),
                    json.dumps(req.end.summary),
                    req.end.id,
                ),
            )
            conn.commit()
        return tsi.CallEndRes()

    def call_read(self, req: tsi.CallReadReq) -> tsi.CallReadRes:
        calls = self.calls_query(
            tsi.CallsQueryReq(
                project_id=req.project_id,
                limit=1,
                filter=tsi.CallsFilter(call_ids=[req.id]),
            )
        ).calls
        return tsi.CallReadRes(call=calls[0] if calls else None)

    def calls_query(self, req: tsi.CallsQueryReq) -> tsi.CallsQueryRes:
        print("REQ", req)
        conn, cursor = get_conn_cursor(self.db_path)
        conds = []
        filter = req.filter
        if filter:
            if filter.op_names:
                assert_parameter_length_less_than_max("op_names", len(filter.op_names))
                or_conditions: list[str] = []

                non_wildcarded_names: list[str] = []
                wildcarded_names: list[str] = []
                for name in filter.op_names:
                    if name.endswith(WILDCARD_ARTIFACT_VERSION_AND_PATH):
                        wildcarded_names.append(name)
                    else:
                        non_wildcarded_names.append(name)

                if non_wildcarded_names:
                    in_expr = ", ".join(f"'{x}'" for x in non_wildcarded_names)
                    or_conditions += [f"op_name IN ({', '.join({in_expr})})"]

                for name_ndx, name in enumerate(wildcarded_names):
                    like_name = name[: -len(WILDCARD_ARTIFACT_VERSION_AND_PATH)] + "%"
                    or_conditions.append(f"op_name LIKE '{like_name}'")

                if or_conditions:
                    conds.append("(" + " OR ".join(or_conditions) + ")")

            if filter.input_refs:
                assert_parameter_length_less_than_max(
                    "input_refs", len(filter.input_refs)
                )
                or_conditions = []
                for ref in filter.input_refs:
                    or_conditions.append(f"input_refs LIKE '%{ref}%'")
                conds.append("(" + " OR ".join(or_conditions) + ")")
            if filter.output_refs:
                assert_parameter_length_less_than_max(
                    "output_refs", len(filter.output_refs)
                )
                or_conditions = []
                for ref in filter.output_refs:
                    or_conditions.append(f"output_refs LIKE '%{ref}%'")
                conds.append("(" + " OR ".join(or_conditions) + ")")
            if filter.parent_ids:
                assert_parameter_length_less_than_max(
                    "parent_ids", len(filter.parent_ids)
                )
                in_expr = ", ".join(f"'{x}'" for x in filter.parent_ids)
                conds += [f"parent_id IN ({in_expr})"]
            if filter.trace_ids:
                assert_parameter_length_less_than_max(
                    "trace_ids", len(filter.trace_ids)
                )
                in_expr = ", ".join(f"'{x}'" for x in filter.trace_ids)
                conds += [f"trace_id IN ({in_expr})"]
            if filter.call_ids:
                assert_parameter_length_less_than_max("call_ids", len(filter.call_ids))
                in_expr = ", ".join(f"'{x}'" for x in filter.call_ids)
                conds += [f"id IN ({in_expr})"]
            if filter.trace_roots_only:
                conds.append("parent_id IS NULL")
            if filter.wb_run_ids:
                in_expr = ", ".join(f"'{x}'" for x in filter.wb_run_ids)
                conds += [f"wb_run_id IN ({in_expr})"]

        if req.query:
            # This is the mongo-style query
            def process_operation(operation: tsi_query.Operation) -> str:
                cond = None

                if isinstance(operation, tsi_query.AndOperation):
                    lhs_part = process_operand(operation.and_[0])
                    rhs_part = process_operand(operation.and_[1])
                    cond = f"({lhs_part} AND {rhs_part})"
                elif isinstance(operation, tsi_query.OrOperation):
                    if len(operation.or_) == 0:
                        raise ValueError("Empty OR operation")
                    elif len(operation.or_) == 1:
                        return process_operand(operation.or_[0])
                    parts = [process_operand(op) for op in operation.or_]
                    cond = f"({' OR '.join(parts)})"
                elif isinstance(operation, tsi_query.NotOperation):
                    operand_part = process_operand(operation.not_[0])
                    cond = f"(NOT ({operand_part}))"
                elif isinstance(operation, tsi_query.EqOperation):
                    lhs_part = process_operand(operation.eq_[0])
                    rhs_part = process_operand(operation.eq_[1])
                    cond = f"({lhs_part} = {rhs_part})"
                elif isinstance(operation, tsi_query.GtOperation):
                    lhs_part = process_operand(operation.gt_[0])
                    rhs_part = process_operand(operation.gt_[1])
                    cond = f"({lhs_part} > {rhs_part})"
                elif isinstance(operation, tsi_query.GteOperation):
                    lhs_part = process_operand(operation.gte_[0])
                    rhs_part = process_operand(operation.gte_[1])
                    cond = f"({lhs_part} >= {rhs_part})"
                elif isinstance(operation, tsi_query.InOperation):
                    lhs_part = process_operand(operation.in_[0])
                    rhs_part = ",".join(process_operand(op) for op in operation.in_[1])
                    cond = f"({lhs_part} IN ({rhs_part}))"
                elif isinstance(operation, tsi_query.ContainsOperation):
                    lhs_part = process_operand(operation.contains_.input)
                    rhs_part = process_operand(operation.contains_.substr)
                    if operation.contains_.case_insensitive:
                        lhs_part = f"LOWER({lhs_part})"
                        rhs_part = f"LOWER({rhs_part})"
                    cond = f"instr({lhs_part}, {rhs_part})"
                else:
                    raise TypeError(f"Unknown operation type: {operation}")

                return cond

            def process_operand(operand: tsi_query.Operand) -> str:
                if isinstance(operand, tsi_query.LiteralOperation):
                    return json.dumps(operand.literal_)
                elif isinstance(operand, tsi_query.GetFieldOperator):
                    field = _transform_external_calls_field_to_internal_calls_field(
                        operand.get_field_, None
                    )
                    return field
                elif isinstance(operand, tsi_query.ConvertOperation):
                    field = process_operand(operand.convert_.input)
                    convert_to = operand.convert_.to
                    if convert_to == "int":
                        sql_type = "INT"
                    elif convert_to == "double":
                        sql_type = "FLOAT"
                    elif convert_to == "bool":
                        sql_type = "BOOL"
                    elif convert_to == "string":
                        sql_type = "TEXT"
                    else:
                        raise ValueError(f"Unknown cast: {convert_to}")
                    return f"CAST({field} AS {sql_type})"
                elif isinstance(
                    operand,
                    (
                        tsi_query.AndOperation,
                        tsi_query.OrOperation,
                        tsi_query.NotOperation,
                        tsi_query.EqOperation,
                        tsi_query.GtOperation,
                        tsi_query.GteOperation,
                        tsi_query.InOperation,
                        tsi_query.ContainsOperation,
                    ),
                ):
                    return process_operation(operand)
                else:
                    raise TypeError(f"Unknown operand type: {operand}")

            filter_cond = process_operation(req.query.expr_)

            conds.append(filter_cond)

        required_columns = ["id", "trace_id", "project_id", "op_name", "started_at"]
        select_columns = [
            key
            for key in tsi.CallSchema.model_fields.keys()
            if key not in ["storage_size_bytes", "total_storage_size_bytes"]
        ]
        if req.columns:
            # TODO(gst): allow json fields to be selected
            simple_columns = list({x.split(".")[0] for x in req.columns})
            if "summary" in simple_columns or req.include_costs:
                simple_columns += ["ended_at", "exception", "display_name"]

            select_columns = [x for x in simple_columns if x in select_columns]
            # add required columns, preserving requested column order
            select_columns += [
                rcol for rcol in required_columns if rcol not in select_columns
            ]

        select_columns_names = [*select_columns]

        if req.include_storage_size:
            select_columns.append(
                "(COALESCE(length(attributes),0) + COALESCE(length(inputs),0) + COALESCE(length(output),0) + COALESCE(length(summary),0))"
            )
            select_columns_names.append("storage_size_bytes")

        join_clause = ""
        if req.include_total_storage_size:
            select_columns.append(
                """
                CASE
                    WHEN calls.parent_id IS NULL THEN trace_stats.total_storage_size_bytes
                    ELSE NULL
                END
            """
            )
            select_columns_names.append("total_storage_size_bytes")

            join_clause += """
                LEFT JOIN (SELECT
                    calls.trace_id as tid,
                    sum(COALESCE(length(attributes),0) + COALESCE(length(inputs),0) + COALESCE(length(output),0) + COALESCE(length(summary),0)) as total_storage_size_bytes
                FROM calls
                GROUP BY calls.trace_id) as trace_stats ON trace_stats.tid=calls.trace_id
            """

        query = f"SELECT {', '.join(select_columns)} FROM calls {join_clause} WHERE deleted_at IS NULL AND project_id = '{req.project_id}'"

        conditions_part = " AND ".join(conds)

        if conditions_part:
            query += f" AND {conditions_part}"

        # Match the batch server:
        if req.sort_by is None:
            order_by = [("started_at", "asc")]
        elif len(req.sort_by) == 0:
            order_by = None
        else:
            order_by = [(s.field, s.direction) for s in req.sort_by]

        if order_by is not None:
            order_parts = []
            for field, direction in order_by:
                json_path: Optional[str] = None
                if field.startswith("inputs"):
                    field = "inputs" + field[len("inputs") :]
                    if field.startswith("inputs."):
                        json_path = field[len("inputs.") :]
                        field = "inputs"
                elif field.startswith("output"):
                    field = "output" + field[len("output") :]
                    if field.startswith("output."):
                        json_path = field[len("output.") :]
                        field = "output"
                elif field.startswith("attributes"):
                    field = "attributes" + field[len("attributes") :]
                    if field.startswith("attributes."):
                        json_path = field[len("attributes.") :]
                        field = "attributes"
                elif field.startswith("summary"):
                    # Handle special summary fields that are calculated rather than stored directly
                    if field == "summary.weave.status":
                        # Create a CASE expression to properly determine the status
                        field = """
                            CASE
                                WHEN exception IS NOT NULL THEN 'error'
                                WHEN ended_at IS NULL THEN 'running'
                                ELSE 'success'
                            END
                        """
                        json_path = None
                    elif field == "summary.weave.latency_ms":
                        # Calculate latency directly using julianday for millisecond precision
                        field = """
                            CASE
                                WHEN ended_at IS NOT NULL THEN
                                    CAST((julianday(ended_at) - julianday(started_at)) * 86400000 AS INTEGER)
                                ELSE 0
                            END
                        """
                        json_path = None
                    elif field == "summary.weave.trace_name":
                        # Handle trace_name field similar to the ClickHouse implementation
                        # If display_name is present, use that
                        # Otherwise, check if op_name is an object reference and extract the name
                        # If not, just use op_name directly
                        field = """
                            CASE
                                WHEN display_name IS NOT NULL AND display_name != '' THEN display_name
                                WHEN op_name IS NOT NULL AND op_name LIKE 'weave-trace-internal:///%' THEN
                                    SUBSTR(
                                        op_name,
                                        LENGTH(op_name) - INSTR(REVERSE(op_name), '/') + 2,
                                        INSTR(SUBSTR(op_name, LENGTH(op_name) - INSTR(REVERSE(op_name), '/') + 2), ':') - 1
                                    )
                                ELSE op_name
                            END
                        """
                        json_path = None
                    else:
                        field = "summary" + field[len("summary") :]
                        if field.startswith("summary."):
                            json_path = field[len("summary.") :]
                            field = "summary"

                assert direction in [
                    "ASC",
                    "DESC",
                    "asc",
                    "desc",
                ], f"Invalid order_by direction: {direction}"
                if json_path:
                    field = f"json_extract({field}, '{quote_json_path(json_path)}')"
                order_parts.append(f"{field} {direction}")

            order_by_part = ", ".join(order_parts)
            query += f" ORDER BY {order_by_part}"

        limit = req.limit or -1
        if limit:
            query += f" LIMIT {limit}"
        if req.offset:
            if limit is None:
                query += " LIMIT -1"
            query += f" OFFSET {req.offset}"

        print("QUERY", query)

        cursor.execute(query)

        query_result = cursor.fetchall()
        calls = []
        for row in query_result:
            call_dict = dict(zip(select_columns_names, row))
            # convert json dump fields into json
            for json_field in ["attributes", "summary", "inputs", "output"]:
                if call_dict.get(json_field):
                    # load json
                    data = json.loads(call_dict[json_field])
                    # do ref expansion
                    if req.expand_columns:
                        data = self._expand_refs(
                            {json_field: data}, req.expand_columns
                        )[json_field]
                    call_dict[json_field] = data
            # convert empty string display_names to None
            if "display_name" in call_dict:
                call_dict["display_name"] = empty_str_to_none(call_dict["display_name"])
            # fill in derived summary fields
            call_dict["summary"] = make_derived_summary_fields(
                summary=call_dict.get("summary") or {},
                op_name=call_dict["op_name"],
                started_at=datetime.datetime.fromisoformat(call_dict["started_at"]),
                ended_at=(
                    datetime.datetime.fromisoformat(call_dict["ended_at"])
                    if call_dict.get("ended_at")
                    else None
                ),
                exception=call_dict.get("exception"),
                display_name=call_dict.get("display_name"),
            )
            # fill in missing required fields with defaults
            for col, mfield in tsi.CallSchema.model_fields.items():
                if mfield.is_required() and col not in call_dict:
                    if isinstance(mfield.annotation, str):
                        call_dict[col] = ""
                    elif isinstance(
                        mfield.annotation, (datetime.datetime, datetime.date)
                    ):
                        raise ValueError(f"Field '{col}' is required for selection")
                    else:
                        call_dict[col] = {}
            calls.append(call_dict)

        if req.include_feedback:
            feedback_query_req = make_feedback_query_req(req.project_id, calls)
            feedback = self.feedback_query(feedback_query_req)
            hydrate_calls_with_feedback(calls, feedback)

        return tsi.CallsQueryRes(calls=[tsi.CallSchema(**call) for call in calls])

    def _expand_refs(
        self, data: dict[str, Any], expand_columns: list[str]
    ) -> dict[str, Any]:
        """
        Recursively expand refs in the data. Only expand refs if requested in the
        expand_columns list. expand_columns must be sorted by depth, shallowest first.
        """
        cols = sorted(expand_columns, key=lambda x: x.count("."))
        for col in cols:
            val = data.get(col)
            if not val:
                val = get_nested_key(data, col)
                if not val:
                    continue

            if not ri.any_will_be_interpreted_as_ref_str(val):
                continue

            if not isinstance(ri.parse_internal_uri(val), ri.InternalObjectRef):
                continue

            derefed_val = self.refs_read_batch(tsi.RefsReadBatchReq(refs=[val])).vals[0]
            set_nested_key(data, col, derefed_val)

        return data

    def calls_query_stream(self, req: tsi.CallsQueryReq) -> Iterator[tsi.CallSchema]:
        return iter(self.calls_query(req).calls)

    def calls_query_stats(self, req: tsi.CallsQueryStatsReq) -> tsi.CallsQueryStatsRes:
        calls = self.calls_query(
            tsi.CallsQueryReq(
                project_id=req.project_id,
                filter=req.filter,
                query=req.query,
            )
        ).calls
        return tsi.CallsQueryStatsRes(
            count=len(calls),
        )

    def calls_delete(self, req: tsi.CallsDeleteReq) -> tsi.CallsDeleteRes:
        assert_non_null_wb_user_id(req)
        # update row with a deleted_at field set to now
        conn, cursor = get_conn_cursor(self.db_path)
        with self.lock:
            recursive_query = """
                WITH RECURSIVE Descendants AS (
                    SELECT id
                    FROM calls
                    WHERE project_id = ? AND
                        deleted_at IS NULL AND
                        parent_id IN (SELECT id FROM calls WHERE id IN ({}))

                    UNION ALL

                    SELECT c.id
                    FROM calls c
                    JOIN Descendants d ON c.parent_id = d.id
                    WHERE c.deleted_at IS NULL
                )
                SELECT id FROM Descendants;
            """.format(", ".join("?" * len(req.call_ids)))

            params = [req.project_id] + req.call_ids
            cursor.execute(recursive_query, params)
            all_ids = [x[0] for x in cursor.fetchall()] + req.call_ids

            # set deleted_at for all children and parents
            delete_query = """
                UPDATE calls
                SET deleted_at = CURRENT_TIMESTAMP
                WHERE deleted_at is NULL AND
                    id IN ({})
            """.format(", ".join("?" * len(all_ids)))
            print("MUTATION", delete_query)
            cursor.execute(delete_query, all_ids)
            conn.commit()

        return tsi.CallsDeleteRes()

    def call_update(self, req: tsi.CallUpdateReq) -> tsi.CallUpdateRes:
        assert_non_null_wb_user_id(req)
        if req.display_name is None:
            raise ValueError("One of [display_name] is required for call update")

        conn, cursor = get_conn_cursor(self.db_path)
        with self.lock:
            cursor.execute(
                "UPDATE calls SET display_name = ? WHERE id = ?",
                (req.display_name, req.call_id),
            )
            conn.commit()
        return tsi.CallUpdateRes()

    def op_create(self, req: tsi.OpCreateReq) -> tsi.OpCreateRes:
        raise NotImplementedError()

    def op_read(self, req: tsi.OpReadReq) -> tsi.OpReadRes:
        raise NotImplementedError()

    def ops_query(self, req: tsi.OpQueryReq) -> tsi.OpQueryRes:
        raise NotImplementedError()

    def obj_create(self, req: tsi.ObjCreateReq) -> tsi.ObjCreateRes:
        conn, cursor = get_conn_cursor(self.db_path)

        processed_result = process_incoming_object_val(
            req.obj.val, req.obj.builtin_object_class
        )
        processed_val = processed_result["val"]
        json_val = json.dumps(processed_val)
        digest = str_digest(json_val)
        project_id, object_id, wb_user_id = (
            req.obj.project_id,
            req.obj.object_id,
            req.obj.wb_user_id,
        )

        # Validate
        object_id_validator(object_id)

        with self.lock:
            if self._obj_exists(cursor, project_id, object_id, digest):
                return tsi.ObjCreateRes(digest=digest)

            cursor.execute("BEGIN TRANSACTION")
            self._mark_existing_objects_as_not_latest(cursor, project_id, object_id)
            version_index = self._get_obj_version_index(cursor, project_id, object_id)
            cursor.execute(
                """INSERT OR IGNORE INTO objects (
                    project_id,
                    object_id,
                    created_at,
                    kind,
                    base_object_class,
                    refs,
                    val_dump,
                    digest,
                    version_index,
                    is_latest,
                    deleted_at,
                    wb_user_id
                ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                ON CONFLICT(project_id, kind, object_id, digest) DO UPDATE SET
                    created_at = excluded.created_at,
                    kind = excluded.kind,
                    base_object_class = excluded.base_object_class,
                    refs = excluded.refs,
                    val_dump = excluded.val_dump,
                    version_index = excluded.version_index,
                    is_latest = excluded.is_latest,
                    deleted_at = excluded.deleted_at
                """,
                (
                    project_id,
                    object_id,
                    datetime.datetime.now().isoformat(),
                    get_kind(processed_val),
                    processed_result["base_object_class"],
                    json.dumps([]),
                    json_val,
                    digest,
                    version_index,
                    1,
                    None,
                    wb_user_id,
                ),
            )
            conn.commit()
        return tsi.ObjCreateRes(digest=digest)

    def _obj_exists(
        self, cursor: sqlite3.Cursor, project_id: str, object_id: str, digest: str
    ) -> bool:
        cursor.execute(
            "SELECT COUNT(*) FROM objects WHERE project_id = ? AND object_id = ? AND digest = ? AND deleted_at IS NULL",
            (project_id, object_id, digest),
        )
        return_row = cursor.fetchone()
        if return_row is None:
            return False
        return return_row[0] > 0

    def _mark_existing_objects_as_not_latest(
        self, cursor: sqlite3.Cursor, project_id: str, object_id: str
    ) -> None:
        """Mark all existing objects with such id as not latest.
        We are creating a new object with the same id, all existing ones are no longer latest.
        """
        cursor.execute(
            "UPDATE objects SET is_latest = 0 WHERE project_id = ? AND object_id = ?",
            (project_id, object_id),
        )

    def _get_obj_version_index(
        self, cursor: sqlite3.Cursor, project_id: str, object_id: str
    ) -> int:
        """Get the version index for a new object with such id."""
        cursor.execute(
            "SELECT COUNT(*) FROM objects WHERE project_id = ? AND object_id = ?",
            (project_id, object_id),
        )
        return_row = cursor.fetchone()
        if return_row is None:
            return 0
        return return_row[0]

    @staticmethod
    def _make_digest_condition(digest: str) -> str:
        if digest == "latest":
            return "is_latest = 1"
        else:
            (is_version, version_index) = digest_is_version_like(digest)
            if is_version:
                return f"version_index = '{version_index}'"
            else:
                return f"digest = '{digest}'"

    def obj_read(self, req: tsi.ObjReadReq) -> tsi.ObjReadRes:
        conds = [f"object_id = '{req.object_id}'"]
        digest_condition = self._make_digest_condition(req.digest)
        conds.append(digest_condition)
        objs = self._select_objs_query(
            req.project_id,
            conditions=conds,
            include_deleted=True,
            metadata_only=req.metadata_only,
        )
        if len(objs) == 0:
            raise NotFoundError(f"Obj {req.object_id}:{req.digest} not found")
        if objs[0].deleted_at is not None:
            raise ObjectDeletedError(
                f"{req.object_id}:v{objs[0].version_index} was deleted at {objs[0].deleted_at}",
                deleted_at=objs[0].deleted_at,
            )
        return tsi.ObjReadRes(obj=objs[0])

    def objs_query(self, req: tsi.ObjQueryReq) -> tsi.ObjQueryRes:
        conds: list[str] = []
        parameters: dict[str, Any] = {}
        if req.filter:
            if req.filter.is_op is not None:
                if req.filter.is_op:
                    conds.append("kind = 'op'")
                else:
                    conds.append("kind != 'op'")
            if req.filter.object_ids:
                placeholders = ",".join(["?" for _ in req.filter.object_ids])
                conds.append(f"object_id IN ({placeholders})")
                parameters["object_ids"] = req.filter.object_ids
            if req.filter.latest_only:
                conds.append("is_latest = 1")
            if req.filter.base_object_classes:
                placeholders = ",".join(["?" for _ in req.filter.base_object_classes])
                conds.append(f"base_object_class IN ({placeholders})")
                parameters["base_object_classes"] = req.filter.base_object_classes

        objs = self._select_objs_query(
            req.project_id,
            conditions=conds,
            parameters=parameters,
            metadata_only=req.metadata_only,
            limit=req.limit,
            offset=req.offset,
            sort_by=req.sort_by,
        )

        return tsi.ObjQueryRes(objs=objs)

    def obj_delete(self, req: tsi.ObjDeleteReq) -> tsi.ObjDeleteRes:
        MAX_OBJECTS_TO_DELETE = 100
        if req.digests and len(req.digests) > MAX_OBJECTS_TO_DELETE:
            raise ValueError(
                f"Object delete request contains {len(req.digests)} objects. Please delete {MAX_OBJECTS_TO_DELETE} or fewer objects at a time."
            )

        # First, select the objects that match the query
        select_query = """
            SELECT digest FROM objects
            WHERE project_id = ? AND
                object_id = ? AND
                deleted_at IS NULL
        """
        parameters = [req.project_id, req.object_id]
        if req.digests:
            digest_conditions = [
                self._make_digest_condition(digest) for digest in req.digests
            ]
            digest_conditions_str = " OR ".join(digest_conditions)
            select_query += f"AND ({digest_conditions_str})"

        conn, cursor = get_conn_cursor(self.db_path)
        cursor.execute(select_query, parameters)
        matching_objects = cursor.fetchall()

        if len(matching_objects) == 0:
            raise NotFoundError(
                f"Object {req.object_id} ({req.digests}) not found when deleting."
            )
        found_digests = {obj[0] for obj in matching_objects}
        if req.digests:
            given_digests = set(req.digests)
            if len(given_digests) != len(found_digests):
                raise NotFoundError(
                    f"Delete request contains {len(req.digests)} digests, but found {len(found_digests)} objects to delete. Diff digests: {given_digests - found_digests}"
                )

        # Create a delete query that will set the deleted_at field to now
        delete_query = """
            UPDATE objects SET deleted_at = CURRENT_TIMESTAMP
            WHERE project_id = ? AND
                object_id = ? AND
                digest IN ({})
        """.format(", ".join("?" * len(found_digests)))
        delete_parameters = [req.project_id, req.object_id] + list(found_digests)

        with self.lock:
            cursor.execute("BEGIN TRANSACTION")
            cursor.execute(delete_query, delete_parameters)
            conn.commit()

        return tsi.ObjDeleteRes(num_deleted=len(matching_objects))

    def table_create(self, req: tsi.TableCreateReq) -> tsi.TableCreateRes:
        conn, cursor = get_conn_cursor(self.db_path)
        insert_rows = []
        for r in req.table.rows:
            if not isinstance(r, dict):
                raise TypeError("All rows must be dictionaries")
            row_json = json.dumps(r)
            row_digest = str_digest(row_json)
            insert_rows.append((req.table.project_id, row_digest, row_json))
        with self.lock:
            cursor.executemany(
                "INSERT OR IGNORE INTO table_rows (project_id, digest, val) VALUES (?, ?, ?)",
                insert_rows,
            )

            row_digests = [r[1] for r in insert_rows]

            table_hasher = hashlib.sha256()
            for row_digest in row_digests:
                table_hasher.update(row_digest.encode())
            digest = table_hasher.hexdigest()

            cursor.execute(
                "INSERT OR IGNORE INTO tables (project_id, digest, row_digests) VALUES (?, ?, ?)",
                (req.table.project_id, digest, json.dumps(row_digests)),
            )
            conn.commit()

        return tsi.TableCreateRes(digest=digest, row_digests=row_digests)

    def table_update(self, req: tsi.TableUpdateReq) -> tsi.TableUpdateRes:
        conn, cursor = get_conn_cursor(self.db_path)
        # conds = ["project_id = {project_id: String}"]

        cursor.execute(
            """
            SELECT
                tables.row_digests
            FROM
                tables
            WHERE
                tables.project_id = ? AND
                tables.digest = ?
            """,
            (req.project_id, req.base_digest),
        )
        query_result = cursor.fetchall()
        final_row_digests: list[str] = json.loads(query_result[0][0])
        new_rows_needed_to_insert = []
        known_digests = set(final_row_digests)

        def add_new_row_needed_to_insert(row_data: Any) -> str:
            if not isinstance(row_data, dict):
                raise TypeError("All rows must be dictionaries")
            row_json = json.dumps(row_data)
            row_digest = str_digest(row_json)
            if row_digest not in known_digests:
                new_rows_needed_to_insert.append((req.project_id, row_digest, row_json))
                known_digests.add(row_digest)
            return row_digest

        updated_digests = []
        for update in req.updates:
            if isinstance(update, tsi.TableAppendSpec):
                new_digest = add_new_row_needed_to_insert(update.append.row)
                final_row_digests.append(new_digest)
                updated_digests.append(new_digest)
            elif isinstance(update, tsi.TablePopSpec):
                if update.pop.index >= len(final_row_digests) or update.pop.index < 0:
                    raise ValueError("Index out of range")
                popped_digest = final_row_digests.pop(update.pop.index)
                updated_digests.append(popped_digest)
            elif isinstance(update, tsi.TableInsertSpec):
                if (
                    update.insert.index > len(final_row_digests)
                    or update.insert.index < 0
                ):
                    raise ValueError("Index out of range")
                new_digest = add_new_row_needed_to_insert(update.insert.row)
                final_row_digests.insert(update.insert.index, new_digest)
                updated_digests.append(new_digest)
            else:
                raise TypeError("Unrecognized update", update)

        # Perform the actual DB inserts
        with self.lock:
            cursor.executemany(
                "INSERT OR IGNORE INTO table_rows (project_id, digest, val) VALUES (?, ?, ?)",
                new_rows_needed_to_insert,
            )

            table_hasher = hashlib.sha256()
            for row_digest in final_row_digests:
                table_hasher.update(row_digest.encode())
            digest = table_hasher.hexdigest()

            cursor.execute(
                "INSERT OR IGNORE INTO tables (project_id, digest, row_digests) VALUES (?, ?, ?)",
                (req.project_id, digest, json.dumps(final_row_digests)),
            )
            conn.commit()

        return tsi.TableUpdateRes(digest=digest, updated_row_digests=updated_digests)

    def table_query(self, req: tsi.TableQueryReq) -> tsi.TableQueryRes:
        conds = []
        parameters: list[str] = []
        if req.filter:
            if req.filter.row_digests:
                conds.append(
                    "tr.digest IN ({})".format(
                        ",".join("?" * len(req.filter.row_digests))
                    )
                )
                parameters.extend(req.filter.row_digests)
            else:
                conds.append("1 = 1")
        else:
            conds.append("1 = 1")

        predicate = " AND ".join(conds)

        # Construct the ORDER BY clause
        order_by = ""
        if req.sort_by:
            sort_parts = []
            for sort in req.sort_by:
                field = sort.field
                direction = sort.direction.upper()
                if "." in field:
                    # Handle nested fields
                    parts = field.split(".")
                    field = f"json_extract(tr.val, '$.{'.'.join(parts)}')"
                else:
                    field = f"json_extract(tr.val, '$.{field}')"
                sort_parts.append(f"{field} {direction}")
            order_by = f"ORDER BY {', '.join(sort_parts)}"
        else:
            order_by = "ORDER BY OrderedDigests.original_order"
        # First get the row IDs by querying tables
        query = f"""
        WITH OrderedDigests AS (
            SELECT
                json_each.value AS digest,
                CAST(json_each.key AS INTEGER) AS original_order
            FROM
                tables,
                json_each(tables.row_digests)
            WHERE
                tables.project_id = ? AND
                tables.digest = ?
        )
        SELECT
            tr.digest,
            tr.val,
            OrderedDigests.original_order
        FROM
            OrderedDigests
            JOIN table_rows tr ON OrderedDigests.digest = tr.digest
        WHERE {predicate}
        {order_by}
        """

        if req.limit is not None:
            query += f" LIMIT {req.limit}"
        if req.offset is not None:
            if req.limit is None:
                query += " LIMIT -1"
            query += f" OFFSET {req.offset}"

        conn, cursor = get_conn_cursor(self.db_path)
        cursor.execute(query, [req.project_id, req.digest] + list(parameters))
        query_result = cursor.fetchall()

        return tsi.TableQueryRes(
            rows=[
                tsi.TableRowSchema(
                    digest=r[0],
                    val=json.loads(r[1]),
                    original_index=r[2],  # Add the original index
                )
                for r in query_result
            ]
        )

    def table_query_stats(self, req: tsi.TableQueryStatsReq) -> tsi.TableQueryStatsRes:
        parameters: list[Any] = [req.project_id, req.digest]

        query = """
        SELECT json_array_length(row_digests)
        FROM
            tables
        WHERE
            tables.project_id = ? AND
            tables.digest = ?
        """

        conn, cursor = get_conn_cursor(self.db_path)
        cursor.execute(query, parameters)
        row = cursor.fetchone()
        count = 0
        if row is not None:
            count = row[0]

        return tsi.TableQueryStatsRes(count=count)

    def refs_read_batch(self, req: tsi.RefsReadBatchReq) -> tsi.RefsReadBatchRes:
        # TODO: This reads one ref at a time, it should read them in batches
        # where it can. Like it should group by object that we need to read.
        # And it should also batch into table refs (like when we are reading a bunch
        # of rows from a single Dataset)
        if len(req.refs) > 1000:
            raise ValueError("Too many refs")

        parsed_refs = [ri.parse_internal_uri(r) for r in req.refs]
        if any(isinstance(r, ri.InternalTableRef) for r in parsed_refs):
            raise ValueError("Table refs not supported")
        parsed_obj_refs = cast(list[ri.InternalObjectRef], parsed_refs)

        def read_ref(r: ri.InternalObjectRef) -> Any:
            conds = [
                f"object_id = '{r.name}'",
                f"digest = '{r.version}'",
            ]
            objs = self._select_objs_query(
                r.project_id,
                conditions=conds,
                include_deleted=True,
            )
            if len(objs) == 0:
                raise NotFoundError(f"Obj {r.name}:{r.version} not found")
            obj = objs[0]
            if obj.deleted_at is not None:
                return None
            val = obj.val
            extra = r.extra
            for extra_index in range(0, len(extra), 2):
                op, arg = extra[extra_index], extra[extra_index + 1]
                if op == ri.DICT_KEY_EDGE_NAME:
                    val = val[arg]
                elif op == ri.OBJECT_ATTR_EDGE_NAME:
                    val = val[arg]
                elif op == ri.LIST_INDEX_EDGE_NAME:
                    val = val[int(arg)]
                elif op == ri.TABLE_ROW_ID_EDGE_NAME:
                    weave_internal_prefix = ri.WEAVE_INTERNAL_SCHEME + ":///"
                    if isinstance(val, str) and val.startswith(weave_internal_prefix):
                        table_ref = ri.parse_internal_uri(val)
                        if not isinstance(table_ref, ri.InternalTableRef):
                            raise ValueError(
                                "invalid data layout encountered, expected TableRef when resolving id"
                            )
                        row = self._table_row_read(
                            project_id=table_ref.project_id,
                            row_digest=arg,
                        )
                        val = row.val
                    else:
                        raise ValueError(
                            "invalid data layout encountered, expected TableRef when resolving id"
                        )
                else:
                    raise ValueError(f"Unknown ref type: {extra[extra_index]}")
            return val

        return tsi.RefsReadBatchRes(vals=[read_ref(r) for r in parsed_obj_refs])

    def feedback_create(self, req: tsi.FeedbackCreateReq) -> tsi.FeedbackCreateRes:
        assert_non_null_wb_user_id(req)
        validate_feedback_create_req(req, self)

        # Augment emoji with alias.
        res_payload = {}
        if req.feedback_type == "wandb.reaction.1":
            em = req.payload["emoji"]
            if emoji.emoji_count(em) != 1:
                raise InvalidRequest(
                    "Value of emoji key in payload must be exactly one emoji"
                )
            req.payload["alias"] = emoji.demojize(em)
            detoned = detone_emojis(em)
            req.payload["detoned"] = detoned
            req.payload["detoned_alias"] = emoji.demojize(detoned)
            res_payload = req.payload

        feedback_id = generate_id()
        created_at = datetime.datetime.now(ZoneInfo("UTC"))
        # TODO: Any validation on weave_ref?
        payload = json.dumps(req.payload)
        MAX_PAYLOAD = 1024
        if len(payload) > MAX_PAYLOAD:
            raise InvalidRequest("Feedback payload too large")
        row: Row = {
            "id": feedback_id,
            "project_id": req.project_id,
            "weave_ref": req.weave_ref,
            "wb_user_id": req.wb_user_id,
            "creator": req.creator,
            "feedback_type": req.feedback_type,
            "payload": req.payload,
            "created_at": created_at,
            "annotation_ref": req.annotation_ref,
            "runnable_ref": req.runnable_ref,
            "call_ref": req.call_ref,
            "trigger_ref": req.trigger_ref,
        }
        conn, cursor = get_conn_cursor(self.db_path)
        with self.lock:
            prepared = TABLE_FEEDBACK.insert(row).prepare(database_type="sqlite")
            cursor.executemany(prepared.sql, prepared.data)
            conn.commit()
        return tsi.FeedbackCreateRes(
            id=feedback_id,
            created_at=created_at,
            wb_user_id=req.wb_user_id,
            payload=res_payload,
        )

    def feedback_query(self, req: tsi.FeedbackQueryReq) -> tsi.FeedbackQueryRes:
        conn, cursor = get_conn_cursor(self.db_path)
        query = TABLE_FEEDBACK.select()
        query = query.project_id(req.project_id)
        query = query.fields(req.fields)
        query = query.where(req.query)
        query = query.order_by(req.sort_by)
        query = query.limit(req.limit).offset(req.offset)
        prepared = query.prepare(database_type="sqlite")
        r = cursor.execute(prepared.sql, prepared.parameters)
        result = TABLE_FEEDBACK.tuples_to_rows(r.fetchall(), prepared.fields)
        return tsi.FeedbackQueryRes(result=result)

    def feedback_purge(self, req: tsi.FeedbackPurgeReq) -> tsi.FeedbackPurgeRes:
        # TODO: Instead of passing conditions to DELETE FROM,
        #       should we select matching ids, and then DELETE FROM WHERE id IN (...)?
        #       This would allow us to return the number of rows deleted, and complain
        #       if too many things would be deleted.
        validate_feedback_purge_req(req)
        conn, cursor = get_conn_cursor(self.db_path)
        query = TABLE_FEEDBACK.purge()
        query = query.project_id(req.project_id)
        query = query.where(req.query)
        prepared = query.prepare(database_type="sqlite")
        with self.lock:
            cursor.execute(prepared.sql, prepared.parameters)
            conn.commit()
        return tsi.FeedbackPurgeRes()

    def feedback_replace(self, req: tsi.FeedbackReplaceReq) -> tsi.FeedbackReplaceRes:
        purge_request = tsi.FeedbackPurgeReq(
            project_id=req.project_id,
            query={
                "$expr": {
                    "$eq": [
                        {"$getField": "id"},
                        {"$literal": req.feedback_id},
                    ],
                }
            },
        )
        self.feedback_purge(purge_request)
        create_req = tsi.FeedbackCreateReq(**req.model_dump(exclude={"feedback_id"}))
        create_result = self.feedback_create(create_req)

        return tsi.FeedbackReplaceRes(
            id=create_result.id,
            created_at=create_result.created_at,
            wb_user_id=create_result.wb_user_id,
            payload=create_result.payload,
        )

    def actions_execute_batch(
        self, req: tsi.ActionsExecuteBatchReq
    ) -> tsi.ActionsExecuteBatchRes:
        raise NotImplementedError(
            "actions_execute_batch is not implemented for SQLite trace server"
        )

    def file_create(self, req: tsi.FileCreateReq) -> tsi.FileCreateRes:
        conn, cursor = get_conn_cursor(self.db_path)
        digest = bytes_digest(req.content)
        with self.lock:
            cursor.execute(
                "INSERT OR IGNORE INTO files (project_id, digest, val) VALUES (?, ?, ?)",
                (
                    req.project_id,
                    digest,
                    req.content,
                ),
            )
            conn.commit()
        return tsi.FileCreateRes(digest=digest)

    def file_content_read(self, req: tsi.FileContentReadReq) -> tsi.FileContentReadRes:
        conn, cursor = get_conn_cursor(self.db_path)
        cursor.execute(
            "SELECT val FROM files WHERE project_id = ? AND digest = ?",
            (req.project_id, req.digest),
        )
        query_result = cursor.fetchone()
        if query_result is None:
            raise NotFoundError(f"File {req.digest} not found")
        return tsi.FileContentReadRes(content=query_result[0])

    def cost_create(self, req: tsi.CostCreateReq) -> tsi.CostCreateRes:
        print("COST CREATE is not implemented for local sqlite", req)
        return tsi.CostCreateRes()

    def cost_query(self, req: tsi.CostQueryReq) -> tsi.CostQueryRes:
        print("COST QUERY is not implemented for local sqlite", req)
        return tsi.CostQueryRes()

    def cost_purge(self, req: tsi.CostPurgeReq) -> tsi.CostPurgeRes:
        print("COST PURGE is not implemented for local sqlite", req)
        return tsi.CostPurgeRes()

    def completions_create(
        self, req: tsi.CompletionsCreateReq
    ) -> tsi.CompletionsCreateRes:
        print("COMPLETIONS CREATE is not implemented for local sqlite", req)
        return tsi.CompletionsCreateRes()

    def otel_export(self, req: tsi.OtelExportReq) -> tsi.OtelExportRes:
        if not isinstance(req.traces, ExportTraceServiceRequest):
            raise TypeError(
                "Expected traces as ExportTraceServiceRequest, got {type(req.traces)}"
            )

        traces_data = [
            ResourceSpans.from_proto(span) for span in req.traces.resource_spans
        ]

        calls = []
        for resource_spans in traces_data:
            for scope_spans in resource_spans.scope_spans:
                for span in scope_spans.spans:
                    start_call, end_call = span.to_call(req.project_id)
                    calls.extend(
                        [
                            {
                                "mode": "start",
                                "req": tsi.CallStartReq(start=start_call),
                            },
                            {"mode": "end", "req": tsi.CallEndReq(end=end_call)},
                        ]
                    )
        res = self.call_start_batch(tsi.CallCreateBatchReq(batch=calls))
        # Return the empty ExportTraceServiceResponse as per the OTLP spec
        return tsi.OtelExportRes()

    def _table_row_read(self, project_id: str, row_digest: str) -> tsi.TableRowSchema:
        conn, cursor = get_conn_cursor(self.db_path)
        # Now get the rows
        cursor.execute(
            """
            SELECT digest, val FROM table_rows
            WHERE project_id = ? AND digest = ?
            """,
            [project_id, row_digest],
        )
        query_result = cursor.fetchone()
        if query_result is None:
            raise NotFoundError(f"Row {row_digest} not found")
        return tsi.TableRowSchema(
            digest=query_result[0], val=json.loads(query_result[1])
        )

    def _select_objs_query(
        self,
        project_id: str,
        conditions: Optional[list[str]] = None,
        parameters: Optional[dict[str, Any]] = None,
        metadata_only: Optional[bool] = False,
        limit: Optional[int] = None,
        include_deleted: bool = False,
        offset: Optional[int] = None,
        sort_by: Optional[list[tsi.SortBy]] = None,
    ) -> list[tsi.ObjSchema]:
        conn, cursor = get_conn_cursor(self.db_path)
        conditions = conditions or []
        if not include_deleted:
            conditions.append("deleted_at IS NULL")
        if not conditions:
            conditions.append("1 = 1")
        pred = " AND ".join(conditions)
        val_dump_part = "'{}' as val_dump" if metadata_only else "val_dump"
        query = f"""
            SELECT
                project_id,
                object_id,
                created_at,
                kind,
                base_object_class,
                {val_dump_part},
                digest,
                version_index,
                is_latest,
                deleted_at,
                wb_user_id
            FROM objects
            WHERE project_id = ? AND {pred}
        """

        if sort_by:
            valid_sort_fields = {"object_id", "created_at"}
            sort_clauses = []
            for sort in sort_by:
                if sort.field in valid_sort_fields and sort.direction in {
                    "asc",
                    "desc",
                }:
                    sort_clauses.append(f"{sort.field} {sort.direction.upper()}")
            if sort_clauses:
                query += f" ORDER BY {', '.join(sort_clauses)}"
        else:
            query += " ORDER BY created_at ASC"

        if limit is not None:
            query += f" LIMIT {limit}"
        if offset is not None:
            query += f" OFFSET {offset}"

        params = [project_id]
        if parameters:
            for param_list in parameters.values():
                params.extend(param_list)

        cursor.execute(query, params)
        query_result = cursor.fetchall()
        result: list[tsi.ObjSchema] = []
        for row in query_result:
            result.append(
                tsi.ObjSchema(
                    project_id=f"{row[0]}",
                    object_id=row[1],
                    created_at=row[2],
                    kind=row[3],
                    base_object_class=row[4],
                    val=json.loads(row[5]),
                    digest=row[6],
                    version_index=row[7],
                    is_latest=row[8],
                    deleted_at=row[9],
                    wb_user_id=row[10],
                )
            )
        return result

    def table_query_stream(
        self, req: tsi.TableQueryReq
    ) -> Iterator[tsi.TableRowSchema]:
        results = self.table_query(req)
        yield from results.rows


def get_type(val: Any) -> str:
    if val == None:
        return "none"
    elif isinstance(val, dict):
        if "_type" in val:
            if "weave_type" in val:
                return val["weave_type"]["type"]
            return val["_type"]
        return "dict"
    elif isinstance(val, list):
        return "list"
    return "unknown"


def get_kind(val: Any) -> str:
    val_type = get_type(val)
    if val_type == "Op":
        return "op"
    return "object"


def _transform_external_calls_field_to_internal_calls_field(
    field: str,
    cast: Optional[str] = None,
) -> str:
    json_path = None
    if field == "inputs" or field.startswith("inputs."):
        if field == "inputs":
            json_path = "$"
        else:
            json_path = quote_json_path(field[len("inputs.") :])
        field = "inputs"
    elif field == "output" or field.startswith("output."):
        if field == "output":
            json_path = "$"
        else:
            json_path = quote_json_path(field[len("output.") :])
        field = "output"
    elif field == "attributes" or field.startswith("attributes."):
        if field == "attributes":
            json_path = "$"
        else:
            json_path = quote_json_path(field[len("attributes.") :])
        field = "attributes"
    elif field == "summary" or field.startswith("summary."):
        if field == "summary":
            json_path = "$"
        else:
            json_path = quote_json_path(field[len("summary.") :])
        field = "summary"

    if json_path is not None:
        sql_type = "TEXT"
        if cast is not None:
            if cast == "int":
                sql_type = "INT"
            elif cast == "float":
                sql_type = "FLOAT"
            elif cast == "bool":
                sql_type = "BOOL"
            elif cast == "str":
                sql_type = "TEXT"
            else:
                raise ValueError(f"Unknown cast: {cast}")
        field = (
            "CAST(json_extract("
            + json.dumps(field)
            + ", '"
            + json_path
            + "') AS "
            + sql_type
            + ")"
        )

    return field



import json
from abc import abstractmethod
from collections.abc import Iterable
from dataclasses import field
from datetime import datetime
from enum import Enum
from typing import Any, Optional, Union
from uuid import UUID

import openinference.semconv.trace as oi
import opentelemetry.semconv_ai as ot
from opentelemetry.proto.common.v1.common_pb2 import AnyValue, KeyValue

from weave.trace_server.trace_server_interface import LLMUsageSchema, SummaryInsertMap


def to_json_serializable(value: Any) -> Any:
    """
    Transform common data types into JSON-serializable values.

    Args:
        value: Any value that needs to be converted to JSON-serializable format

    Returns:
        A JSON-serializable version of the input value

    Raises:
        ValueError: If the value type is not supported for JSON serialization
    """
    if value is None:
        return None
    elif isinstance(value, (str, int, float, bool)):
        return value
    elif isinstance(value, (list, tuple)):
        return [to_json_serializable(item) for item in value]
    elif isinstance(value, dict):
        return {str(k): to_json_serializable(v) for k, v in value.items()}
    elif isinstance(value, datetime):
        return value.isoformat()
    elif isinstance(value, UUID):
        return str(value)
    elif isinstance(value, Enum):
        return value.value
    elif hasattr(value, "__dataclass_fields__"):  # Handle dataclasses
        return {
            k: to_json_serializable(getattr(value, k))
            for k in value.__dataclass_fields__
        }
    else:
        raise ValueError(f"Unsupported type for JSON serialization: {type(value)}")


def resolve_pb_any_value(value: AnyValue) -> Any:
    """
    Resolve the value field of an AnyValue protobuf message.

    Args:
        value: An AnyValue protobuf message

    Returns:
        The resolved value as a Python type

    Raises:
        ValueError: If the AnyValue has no value set or has an unsupported type
    """
    if value.HasField("string_value"):
        return value.string_value
    elif value.HasField("bool_value"):
        return value.bool_value
    elif value.HasField("int_value"):
        return value.int_value
    elif value.HasField("double_value"):
        return value.double_value
    elif value.HasField("array_value"):
        return [resolve_pb_any_value(v) for v in value.array_value.values]
    elif value.HasField("kvlist_value"):
        return dict(resolve_pb_key_value(kv) for kv in value.kvlist_value.values)
    elif value.HasField("bytes_value"):
        return value.bytes_value
    else:
        raise ValueError("AnyValue has no value set")


def resolve_pb_key_value(key_value: KeyValue) -> tuple[str, Any]:
    """
    Resolve a KeyValue protobuf message into a tuple of key and value.

    Args:
        key_value: A KeyValue protobuf message

    Returns:
        A tuple containing the key and resolved value

    Raises:
        ValueError: If the KeyValue has no value set or has an unsupported type
    """
    return (key_value.key, resolve_pb_any_value(key_value.value))


def _get_value_from_nested_dict(d: dict[str, Any], key: str) -> Any:
    """Get a value from a nested dictionary using a dot-separated key."""
    if "." not in key:
        return d.get(key)

    parts = key.split(".")
    current = d
    for part in parts:
        if part not in current or not isinstance(current, dict):
            return None
        current = current[part]
    return current


def _set_value_in_nested_dict(d: dict[str, Any], key: str, value: Any) -> None:
    """Set a value in a nested dictionary using a dot-separated key."""
    if "." not in key:
        d[key] = value
        return

    parts = key.split(".")
    current = d
    for _, part in enumerate(parts[:-1]):
        if part not in current:
            current[part] = {}
        current = current[part]
    current[parts[-1]] = value


def convert_numeric_keys_to_list(
    obj: dict[str, Any],
) -> Union[dict[str, Any], list[Any]]:
    """
    Convert dictionaries with numeric-only keys to lists.

    If all keys in a dictionary are numeric strings (0, 1, 2, ...),
    convert it to a list. Recursively processes nested dictionaries.
    """
    # Process all nested dictionaries first
    for key, value in obj.items():
        if isinstance(value, dict):
            obj[key] = convert_numeric_keys_to_list(value)

    # Check if all keys are numeric strings and contiguous starting from 0
    try:
        keys = sorted(int(k) for k in obj.keys())
        if keys == list(range(len(keys))) and all(
            isinstance(k, str) and k.isdigit() for k in obj.keys()
        ):
            # Convert to list, preserving order
            return [obj[str(i)] for i in range(len(keys))]
    except (ValueError, TypeError):
        # Not all keys are numeric
        pass

    return obj


def expand_attributes(
    kv: Iterable[tuple[str, str]], json_attributes: list[str] = []
) -> dict[str, Any]:
    """
    Expand a flattened JSON attributes file into a nested Python dictionary.

    Args:
        file_path: Path to the JSON file with flattened attributes
        json_attributes: list of attributes to parse as JSON strings

    Returns:
        A nested Python dictionary
    """
    # Read the JSON file

    # Create the result dictionary
    result_dict: dict[str, Any] = {}

    # Process each key-value pair
    for flat_key, value in kv:
        # Check if the value should be parsed as JSON
        should_parse_as_json = any(
            flat_key.endswith(attr) or flat_key == attr for attr in json_attributes
        )

        if should_parse_as_json and isinstance(value, str):
            try:
                value = json.loads(value)
            except json.JSONDecodeError:
                # If JSON parsing fails, keep the original value
                pass

        # Add the nested key to the result
        _set_value_in_nested_dict(result_dict, flat_key, value)

    # Convert dictionaries with numeric keys to lists
    return result_dict


def flatten_attributes(
    data: dict[str, Any], json_attributes: list[str] = []
) -> dict[str, Any]:
    """
    Flatten a nested Python dictionary into a flat dictionary with dot-separated keys.

    Args:
        data: Nested Python dictionary to flatten
        json_attributes: list of attributes to stringify as JSON

    Returns:
        A flattened dictionary with dot-separated keys
    """
    result: dict[str, Any] = {}

    def _flatten(obj: Union[dict[str, Any], list[Any]], prefix: str = "") -> None:
        # Check if the entire object should be stringified as JSON
        should_stringify_entire_obj = any(
            prefix.rstrip(".") == attr for attr in json_attributes
        )

        if should_stringify_entire_obj:
            result[prefix.rstrip(".")] = json.dumps(obj)
            return

        if isinstance(obj, dict):
            for key, value in obj.items():
                new_key = f"{prefix}{key}" if prefix else key

                # Check if this exact key's value should be stringified as JSON
                should_stringify_as_json = any(
                    new_key == attr for attr in json_attributes
                )

                if (
                    isinstance(value, dict) or isinstance(value, list)
                ) and not should_stringify_as_json:
                    # Recursively flatten nested dictionaries or lists
                    _flatten(value, f"{new_key}.")
                else:
                    # If the value matches a JSON attribute, stringify it
                    if should_stringify_as_json and not isinstance(value, str):
                        value = json.dumps(value)
                    result[new_key] = value
        elif isinstance(obj, list):
            # Handle lists by using numeric indices as keys
            for i, item in enumerate(obj):
                new_key = f"{prefix}{i}"

                # Check if this exact key's value should be stringified as JSON
                should_stringify_as_json = any(
                    new_key == attr for attr in json_attributes
                )

                if (
                    isinstance(item, dict) or isinstance(item, list)
                ) and not should_stringify_as_json:
                    # Recursively flatten nested dictionaries or lists
                    _flatten(item, f"{new_key}.")
                else:
                    # If the item matches a JSON attribute, stringify it
                    if should_stringify_as_json and not isinstance(item, str):
                        item = json.dumps(item)
                    result[new_key] = item

    _flatten(data)
    return result


def get_attribute(data: dict[str, Any], key: str) -> Any:
    """
    Get the value of a nested attribute from either a nested or flattened dictionary.

    Args:
        data: dictionary to get value from
        key: Dot-separated key to get

    Returns:
        The value at the specified key or None if not found
    """
    # Check if it's a flat dictionary
    if key in data:
        return data[key]

    # Try to get from nested structure
    return _get_value_from_nested_dict(data, key)


def unflatten_key_values(
    key_values: Iterable[KeyValue],
) -> dict[str, Any]:
    """
    Transform a list of KeyValue pairs into a nested dictionary structure.

    Args:
        key_values: An iterable of KeyValue protobuf messages
        separator: The character used to separate nested keys (default: '.')

    Returns:
        A nested dictionary where keys are split by the separator

    Example:
        Input: [("llm.token_count.completion", 123)]
        Output: {"llm": {"token_count": {"completion": 123}}}

        Input: [
            ("retrieval.documents.0.document.content", 'A'),
            ("retrieval.documents.1.document.content", 'B')
        ]
        Output: {
            "retrieval": {
                "documents": {
                    "0": {"document": {"content": "A"}},
                    "1": {"document": {"content": "B"}}
                }
            }
        }
    """
    iterator = ((kv.key, resolve_pb_any_value(kv.value)) for kv in key_values)
    return expand_attributes(iterator, json_attributes=[])


class ConventionType(Enum):
    OPENINFERENCE = "openinference"
    OPENTELEMETRY = "gen_ai"
    CUSTOM = "custom"


class Attributes:
    _attributes: dict[str, Any] = field(default_factory=dict)

    def __init__(self, _attributes: dict[str, Any] = {}) -> None:
        self._attributes = _attributes

    def __getitem__(self, key: str) -> Any:
        return self._attributes.__getitem__(key)

    def __setitem__(self, key: str, value: Any) -> None:
        return self._attributes.__setitem__(key, value)

    def get(self, key: str, default: Any = None) -> Any:
        return self._attributes.get(key, default)

    def get_attribute_value(self, key: str) -> Any:
        return get_attribute(self._attributes, key)

    @abstractmethod
    def get_weave_inputs(self) -> dict[str, Any]: ...

    @abstractmethod
    def get_weave_outputs(self) -> Any: ...

    @abstractmethod
    def get_weave_usage(self) -> LLMUsageSchema: ...

    @abstractmethod
    def get_weave_summary(self) -> SummaryInsertMap: ...

    @abstractmethod
    def get_weave_attributes(self, extra: Optional[dict[str, Any]] = None) -> Any: ...


# If we don't have any conventions to follow, just dump everything to attributes
# Later we may add support for user defined conventions through headers
class GenericAttributes(Attributes):
    def get_weave_usage(self) -> LLMUsageSchema:
        return {}

    def get_weave_summary(self) -> SummaryInsertMap:
        return {}

    def get_weave_inputs(self) -> Any:
        return {}

    def get_weave_outputs(self) -> Any:
        return {}

    def get_weave_attributes(self, extra: Optional[dict[str, Any]] = None) -> Any:
        result = self._attributes.copy()
        result.update(extra or {})
        return result


class OpenInferenceAttributes(Attributes):
    def get_weave_usage(self) -> LLMUsageSchema:
        prompt_tokens = self.get_attribute_value(
            oi.SpanAttributes.LLM_TOKEN_COUNT_PROMPT
        )
        completion_tokens = self.get_attribute_value(
            oi.SpanAttributes.LLM_TOKEN_COUNT_COMPLETION
        )
        total_tokens = self.get_attribute_value(oi.SpanAttributes.LLM_TOKEN_COUNT_TOTAL)
        return LLMUsageSchema(
            prompt_tokens=int(prompt_tokens) if prompt_tokens else None,
            completion_tokens=int(completion_tokens) if completion_tokens else None,
            total_tokens=int(total_tokens) if total_tokens else None,
        )

    def get_weave_summary(self) -> SummaryInsertMap:
        return SummaryInsertMap(usage={"usage": self.get_weave_usage()})

    def get_weave_outputs(self) -> Any:
        outputs: dict[str, Any] = (
            self.get_attribute_value(oi.SpanAttributes.LLM_OUTPUT_MESSAGES) or {}
        )
        result = {}
        for k, v in outputs.items():
            if k.isdigit() and isinstance(v, dict):
                for key in v.keys():
                    result[key + f"_{k}"] = v[key]
        return to_json_serializable(result)

    def get_weave_inputs(self) -> Any:
        inputs: dict[str, Any] = (
            self.get_attribute_value(oi.SpanAttributes.LLM_INPUT_MESSAGES) or {}
        )
        result = {}
        for k, v in inputs.items():
            if k.isdigit() and isinstance(v, dict):
                for key in v.keys():
                    result[key + f"_{k}"] = v[key]
        return to_json_serializable(result)

    def get_weave_attributes(
        self, extra: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        system = self.get_attribute_value(oi.SpanAttributes.LLM_SYSTEM)
        provider = self.get_attribute_value(oi.SpanAttributes.LLM_PROVIDER)
        invocation_parameters = self.get_attribute_value(
            oi.SpanAttributes.LLM_INVOCATION_PARAMETERS
        )
        kind = self.get_attribute_value(oi.SpanAttributes.OPENINFERENCE_SPAN_KIND)
        model = self.get_attribute_value(oi.SpanAttributes.LLM_MODEL_NAME)
        attributes = {
            "system": str(system) if system else None,
            "provider": str(provider) if provider else None,
            "kind": str(kind) if kind else None,
            "model": str(model) if model else None,
        }

        if invocation_parameters:
            try:
                js = json.loads(invocation_parameters)
                for k, v in js.items():
                    attributes[k] = str(v)
            except json.JSONDecodeError:
                raise ValueError(f"Invalid JSON string: {invocation_parameters}")

        if extra:
            attributes.update(extra)

        return to_json_serializable(attributes)


class OpenTelemetryAttributes(Attributes):
    def get_weave_usage(self) -> LLMUsageSchema:
        prompt_tokens = self.get_attribute_value(
            ot.SpanAttributes.LLM_USAGE_PROMPT_TOKENS
        )
        completion_tokens = self.get_attribute_value(
            ot.SpanAttributes.LLM_USAGE_COMPLETION_TOKENS
        )
        total_tokens = self.get_attribute_value(
            ot.SpanAttributes.LLM_USAGE_TOTAL_TOKENS
        )
        return LLMUsageSchema(
            prompt_tokens=int(prompt_tokens) if prompt_tokens else None,
            completion_tokens=int(completion_tokens) if completion_tokens else None,
            total_tokens=int(total_tokens) if total_tokens else None,
        )

    def get_weave_summary(self) -> SummaryInsertMap:
        return SummaryInsertMap(usage={"usage": self.get_weave_usage()})

    def get_weave_outputs(self) -> Any:
        outputs: dict[str, Any] = (
            self.get_attribute_value(ot.SpanAttributes.LLM_COMPLETIONS) or {}
        )
        return to_json_serializable(outputs)

    def get_weave_inputs(self) -> Any:
        inputs: dict[str, Any] = (
            self.get_attribute_value(ot.SpanAttributes.LLM_PROMPTS) or {}
        )
        return to_json_serializable(inputs)

    def get_weave_attributes(
        self, extra: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        max_tokens = self.get_attribute_value(ot.SpanAttributes.LLM_REQUEST_MAX_TOKENS)
        system = self.get_attribute_value(ot.SpanAttributes.LLM_SYSTEM)
        kind = self.get_attribute_value(ot.SpanAttributes.TRACELOOP_SPAN_KIND)
        model = self.get_attribute_value(ot.SpanAttributes.LLM_RESPONSE_MODEL)

        attributes = {
            "system": str(system) if system else None,
            "max_tokens": int(max_tokens) if max_tokens else None,
            "kind": str(kind) if kind else None,
            "model": str(model) if model else None,
        }

        if extra:
            attributes.update(extra)

        return to_json_serializable(attributes)


class AttributesFactory:
    def from_proto(self, key_values: Iterable[KeyValue]) -> "Attributes":
        expanded = unflatten_key_values(key_values)
        if get_attribute(expanded, ConventionType.OPENINFERENCE.value):
            return OpenInferenceAttributes(expanded)
        elif get_attribute(expanded, ConventionType.OPENTELEMETRY.value):
            return OpenTelemetryAttributes(expanded)
        return GenericAttributes(expanded)
